unit Vartbl;
{$I ..\PDXRBLD.INC}

interface
uses Wintypes,WinProcs,Classes,SysUtils,dbiTypes,dbiProcs,DB,DBTables;

Type

  TLocateOption = (loCaseInsensitive, loPartialKey);
  TLocateOptions = set of TLocateOption;

  tVariantType =  (_vtNull,
                   _vtInteger,
                   _vtBoolean,
                   _vtExtended,
                   _vtString,
                   _vtInvalid);

  tVariantRecord = record case vType: tvariantType of
                   _vtNull    :();
                   _vtInteger :(I: integer);
                   _vtBoolean :(B: Boolean);
                   _vtExtended:(E: Extended);
                   _vtString  :(S: string);
                   end;

  varrecarray = array [0..99] of tVariantRecord;

  variant = class
  private
    CntValues: word;
    values: ^varrecarray;
    Procedure SetVarValue(i: word; Value: tVariantRecord);
    Function GetVarValue(i: word): tVariantRecord;
  public
    constructor Create(Cnt: Word);
    destructor Destroy; override;
    property value[i: word]: tVariantRecord read GetVarValue write SetVarValue; default;
  end;

  tTable = class(DBTables.tTable)
  private
    Procedure SetVarValue(const Fldname: string; Value: tVariantRecord);
    Function GetVarValue(const Fldname: string): tVariantRecord;
  public
    function Locate(const KeyFields: string; const KeyValues: variant;
                    const Options: TLocateOptions): Boolean;
    Property Value[const Fldname: String]: tvariantRecord read GetVarValue write SetVarValue; default;
  end;

function VarArrayCreate(const Bounds: array of Integer): Variant;
Procedure VarArrayFree(var v: Variant);
Function EqualVariants(const v1,v2: tVariantRecord): Boolean;

implementation
Uses LowStuff;

Function EqualVariants(const v1,v2: tVariantRecord): Boolean;
begin
  Result := v1.VType=v2.vType;
  if Result then
  case v1.vType of
  _vtNull    : {True};
  _vtInteger : Result := v1.I = v2.I;
  _vtBoolean : Result := v1.B = v2.B;
  _vtExtended: Result := v1.E = v2.E;
  _vtString  : Result := v1.S = v2.S;
  else raise Exception.Create('Invalid variant type');
  end;
end;

function VarArrayCreate(const Bounds: array of Integer): Variant;
var i,x: integer;
    size: word;
begin
  size := 1;
  For i := 0 to High(Bounds) do begin
    if odd(i) then size := size*(bounds[i]-x+1)
    else x := bounds[i]
  end;
  Result := variant.Create(size);
end;

Procedure VarArrayFree(var v: Variant);
begin
  v.Free;
  v := nil;
end;

const
  FiledType2varType : array [tFieldType] of tVariantType =
  (
  {ftUnknown}  _vtInvalid,
  {ftString}   _vtString,
  {ftSmallint} _vtInteger,
  {ftInteger}  _vtInteger,
  {ftWord}     _vtInteger,
  {ftBoolean}  _vtBoolean,
  {ftFloat}    _vtExtended,
  {ftCurrency} _vtExtended,
  {ftBCD}      _vtExtended,
  {ftDate}     _vtExtended,
  {ftTime}     _vtExtended,
  {ftDateTime} _vtExtended,
  {ftBytes}    _vtInvalid,
  {ftVarBytes} _vtInvalid,
  {ftBlob}     _vtInvalid,
  {ftMemo}     _vtInvalid,
  {ftGraphic}  _vtInvalid
  );

procedure VarRec2Field(vRec: tVariantrecord; Fld: tField);
begin
  case vRec.vType of
  _vtNull    : Fld.Clear;
  _vtInteger : Fld.AsInteger := vRec.I;
  _vtBoolean : Fld.AsBoolean := vRec.B;
  _vtExtended: Fld.AsFloat   := vRec.E;
  _vtString  : Fld.AsString  := vRec.S;
  else raise Exception.Create('invalid field type');
  end;
end;
procedure Field2VarRec(Fld: tField; VAR vRec: tVariantrecord);
begin
  if Fld.IsNull then vRec.vType := _vtNull else begin
    vRec.vType := FiledType2varType[Fld.DataType];
    case vRec.vType of
    _vtInteger : vRec.I  := Fld.AsInteger;
    _vtBoolean : vRec.B  := Fld.AsBoolean;
    _vtExtended: vRec.E  := Fld.AsFloat;
    _vtString  : vRec.S  := Fld.AsString;
    else raise Exception.Create('invalid field type');
    end;
  end;
end;

{ variant }

constructor variant.Create(Cnt: Word);
var i: word;
begin
  inherited create;
  CntValues := Cnt;
  GetMem(values,Cnt*SizeOf(tVariantRecord));
  FillChar(values^,Cnt*SizeOf(tVariantRecord),#0);
end;

destructor variant.Destroy;
var i: integer;
begin
  if values <> nil then
    FreeMem(values,CntValues*SizeOf(tVariantRecord));
  inherited destroy;
end;

Procedure variant.SetVarValue(i: word; Value: tVariantRecord);
begin
  values^[i] := value;
end;
Function variant.GetVarValue(i: word): tVariantRecord;
begin
  result := values^[i]
end;



{tTable}

Procedure tTable.SetVarValue;
begin
  VarRec2Field(value,FieldByname(Fldname));
end;

Function tTable.GetVarValue;
begin
  Field2VarRec(FieldByname(Fldname),Result);
end;

function tTable.Locate(const KeyFields: string; const KeyValues: variant;
                     const Options: TLocateOptions): Boolean;
var FieldList: tStringList;
    KeyvaluesHigh: integer;
  function match: boolean;
  var i: integer;
  begin
    result := true;
    for i := 0 to FieldList.Count - 1 do begin
      case Keyvalues[i].vType of
      _vtNull    : if not tField(FieldList.Objects[i]).IsNull
                   then Result := false;
      _vtInteger : if Keyvalues[i].I <> tField(FieldList.Objects[i]).AsInteger
                   then Result := false;
      _vtBoolean : if Keyvalues[i].B <> tField(FieldList.Objects[i]).AsBoolean
                   then Result := false;
      _vtExtended: if Keyvalues[i].E <> tField(FieldList.Objects[i]).AsFloat
                   then Result := false;
      _vtString  : if loCaseInsensitive in Options then begin
                     if loPartialKey in Options then begin
                       if Pos(AnsiUpperCase(Keyvalues[i].S),
                              AnsiUpperCase(tField(FieldList.Objects[i]).AsString)) <> 1
                       then Result := false;
                     end else begin
                       if AnsiUpperCase(Keyvalues[i].S) <>
                          AnsiUpperCase(tField(FieldList.Objects[i]).AsString)
                       then Result := false;
                     end;
                   end else
                   if loPartialKey in Options then begin
                     if Pos(Keyvalues[i].S,
                            tField(FieldList.Objects[i]).AsString) <> 1
                     then Result := false;
                   end
                   else
                   if Keyvalues[i].S  <> tField(FieldList.Objects[i]).AsString
                   then Result := false;
      else raise Exception.Create('invalid field type');
      end;
      if not Result then break;
    end
  end;
var i: integer;
begin
  Result := false;
  FieldList:= tStringList.Create;
  try
    parseWords(KeyFields,[';'],FieldList);
    KeyvaluesHigh := Keyvalues.CntValues-1;
    if FieldList.Count <> succ(KeyvaluesHigh)
       then raise Exception.Create('Invalid locate request');
    if FieldList.Count = 0 then exit;
  { set fields objects as Objects of String list }
    For i := 0 to FieldList.Count-1 do
      FieldList.Objects[i] := FieldByname(FieldList[i]);
    First;
    While not Eof do begin
      if match then begin
        Result := true;
        break;
      end;
      next;
    end;
  finally
    FieldList.Free;
  end;
end;


end.
