unit main;

{ Purpose of this simple demo application is to show how to launch Pdxrbld
  from within other applications. Before the database becomes connected
  (that is, before the first table is opened), you can invoke Pdxrbld
  with proper command line. The location of pdxrbld.exe can be determined
  by inspecting registry. Function "ExecWait" (Source: Dmitriy Anisimkov <ts@quadrat.omsk.su>)
  executes Pdxrbld and waits until it finishes. Result returned is the exit code
  of called application - this is set to -1 in Pdxrbld if any errors have been detected.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, DB, DBTables, Grids, DBGrids, DBCtrls, StdCtrls, Registry, RegStr;

type
  TForm1 = class(TForm)
    DataSource1: TDataSource;
    Table1: TTable;
    Panel1: TPanel;
    DBGrid1: TDBGrid;
    DBNavigator1: TDBNavigator;
    btnQuit: TButton;
    btnPack: TButton;
    Database1: TDatabase;
    procedure btnQuitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnPackClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}


{The following code fragment will execute any program, DOS
or 32-bit, and wait until the called program has finished
running. 
----------------------------------------------------------- }
procedure ChkBool(Value: Boolean; const Msg: String);
begin
If (Value = false) then
  raise exception.create(Msg);
end;

{ Source: Dmitriy Anisimkov <ts@quadrat.omsk.su> }
function ExecWait(const Cmd: String): Integer;
var
  ProcessInfo: TProcessInformation;
  hProcess: THandle;
  ReturnCode: Integer;
  StartupInfo: TStartupInfo;
begin
  FillChar(StartupInfo, SizeOf(StartupInfo), 0);
  ChkBool(CreateProcess(nil, PChar(Cmd), nil, nil, False,
          CREATE_DEFAULT_ERROR_MODE + NORMAL_PRIORITY_CLASS,
          NIL, NIL, StartupInfo, ProcessInfo),
          'Error during CreateProcess');
  hProcess := ProcessInfo.hProcess; // save the process handle
  //Close the thread handle as soon as it is no longer needed
  CloseHandle(ProcessInfo.hThread);
  ReturnCode := WaitForSingleObject(hProcess, INFINITE);
  ChkBool(dword(ReturnCode) <> WAIT_FAILED, 'Error in WaitForSingleObject');
  // The process terminated
  ChkBool(GetExitCodeProcess(hProcess, dword(Result)),
          'Error in GetExitCodeProcess');
  // Close the process handle as soon as it is no longer needed
  ChkBool(CloseHandle(hProcess),
          'Error in process Close Handle');
end;


procedure TForm1.btnQuitClick(Sender: TObject);
begin
  close;
end;

function GetRbldpath: string;
{ Checks the registry and returns path to Pdxrbld.exe.
  Returns empty string if either path has not been found
  or path stored in registry is invalid }
{$IFDEF VER90}
    procedure makesinglebackslashes(var s: string);
    var i: integer;
    begin
      repeat
        i := pos('\\',s);
        if i > 0 then System.Delete(s,i,1);
      until i=0
    end;
{$ENDIF}
var REGSTRPATHAPPPATHS: String;
begin
  Result := '';
  REGSTRPATHAPPPATHS:=REGSTR_PATH_APPPATHS;
{$IFDEF VER90} // D2 key constanst have double backslashes
  makesinglebackslashes(REGSTRPATHAPPPATHS);
{$ENDIF}
  with tRegistry.Create do Try
    RootKey := HKEY_LOCAL_MACHINE;
    if OpenKey(REGSTRPATHAPPPATHS+'\Pdxrbld.exe',False) then
      Result := ReadString('');
  finally
    Free;
  end;
  if (Result <> '') and not FileExists(Result) then Result := '';
end;


procedure TForm1.FormCreate(Sender: TObject);
var s: string;
    Rslt: Integer;
begin
{ before opening the first table, check whether Pdxrbld is installed }
  s := GetRbldpath;
  if (s <> '') and { ... yes, it is installed }
    (MessageDlg(Format('Do you want to check %s database?',[Database1.Aliasname]),
                       mtConfirmation,
                       [mbYes,mbNo],0) = mrYes) { ... user confirmed Pdxrbld launch }
  then begin
    { parameter "-R1" tells Pdxrbld to rebuild damaged tables, -Q+ forces "Quiet mode" }
    Rslt := ExecWait(Format('%s -Q+ -R1 "/A%s"',[s,Database1.Aliasname]));
    if Rslt < 0  then { errors have been detected .. }
    case Rslt of
     -2: { database in use } ShowMessage('Cannot perform check while the database is in use');
     else begin { Some other error occurred. Give the user an option to exit gracefully }
        if MessageDlg('Errors have been detected. Do you want to continue?',mtWarning,[mbYes,mbNo],0) <> mrYes
        then begin
          application.terminate;
          exit;
        end;
      end
    end; {case}
  end;
  table1.Open;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  table1.Close;
end;

procedure TForm1.btnPackClick(Sender: TObject);
var s: string;
    Rslt: Integer;
begin
  s := GetRbldpath;
  if s <> '' then begin
    if MessageDlg(Format('Check and pack database %s?',[Database1.Aliasname]),
                mtConfirmation,[mbYes,mbNo],0) = mrYes then begin
      database1.Connected := false; // disconnect database first                
      Try
    { parameter "-P+" tells Pdxrbld to pack tables. -Q+ forces "Quiet mode" }
        Rslt := ExecWait(Format('%s -Q+ -R1 -P+ "/A%s"',[s,Database1.Aliasname]));
        if Rslt < 0  then
        case Rslt of
          -2: { database in use } ShowMessage('Cannot perform check while the database is in use');
         else begin { Some other error occurred }
           ShowMessage('Some errors occurred !');
         end;
        end; {case}
      Finally
        table1.Open; // reopen table
      End;
    end;
  end else raise Exception.Create('Pdxrbld unavailable');
end;

end.
