unit LowStuff;
{$I PDXRBLD.INC}

interface
uses
{$IFDEF WIN32}
  Windows,
  BDE,
  Tutil32,
  Px3PImp,
{$ELSE}
  WinTypes,WinProcs,Ver,
  dbiTypes,dbiProcs,dbiErrs,
  Tutility,
{$ENDIF}
  SysUtils, Classes, FileCtrl, DB, DBTables,
  Translat;

Const
  PdoxUsrsLck  = 'Pdoxusrs.Lck'; { Paradox lock file }
  MaxPasswLen  = 15; { maximum Paradox password length }
{$IFNDEF WIN32}
  max_Path = 176;
{$ENDIF}

type
  tCharSet = Set of Char;
  EDatabaseInUse = class(Exception);

{ Pads string to right }
Function PadRight(const s: string; ToLen: integer):string;

{ Gets Vrsion major (V1), Vesrion minor (V2), Release  version (V3)
  and Build number (V4) from module version info
  (Steve Shafer, Delphi newsgroups), modified }
procedure GetBuildInfo(forModule: tHandle; var V1, V2, V3, V4: Word);

{ Adds backslash to the path if not already present }
Function _ForcebackSlash(const Dir:String):String;

{ Populates List with words from the sentence }
Procedure ParseWords(Const Sentence:string; Separators: tCharSet; List: tStringList);

{ Returns private directory for the app: "Windows temp directory\application name".
  At the same time, sets BDE session.privatedir to the result }

{ Sets session's private directory to "Windows temp directory\application name".
  and returns path to private directory }
function SetPrivDir: string;

{ Checks whether database is in use by attempting to delete Pdoxusrs.Lck file }
Procedure CheckDatabaseInUse(const DBPath: string);

{ Gets DBI error string corresponding to DBI error code }
function GetBDEErrorMsg(rslt: DBIResult): string;

{ Gets TUTILITY resource string corresponding to resource ID}
function GetTUTILITYString(ResID: integer): string;

{$IFNDEF WIN32}
{ Returns path to temporary directoory by querying environment for
  'TEMP' or 'TMP' string. If not found, returns zero, otherwise
  length of the path }
Function GetTempPath(nBufferLength: word; lpBuffer: pChar): word;

{ Extracts drive part of the file path }
function ExtractFileDrive(const FileName: string): string;

{ Extracts directory part of the file path }
function ExtractFileDir(const FileName: string): string;
{$ELSE}
{px3pProtect with result check - exception is raised if Px3pProtect failed }
Procedure MyPx3PProtect(Const Px3PKey: string);
{$ENDIF}

implementation

Function PadRight(const s: string; ToLen: integer):string;
begin
  Result := s;
  while length(Result) < ToLen do Result := Result + ' ';
end;

procedure GetBuildInfo(forModule: tHandle; var V1, V2, V3, V4: Word);
var
  VerInfoSize:{$IFDEF WIN32}DWORD{$ELSE}WORD{$ENDIF};
  VerInfo:Pointer;
  VerValueSize:{$IFDEF WIN32}DWORD{$ELSE}WORD{$ENDIF};
  VerValue:{$IFDEF WIN32}PVSFixedFileInfo{$ELSE}Pvs_FixedFileInfo{$ENDIF};
  Dummy:{$IFDEF WIN32}DWORD{$ELSE}Longint{$ENDIF};
  ForFile: array[0..max_path] of char;
begin
  V1 := 0;
  V2 := 0;
  V3 := 0;
  V4 := 0;
  if forModule = 0 then exit;
  if GetModuleFilename(forModule,@ForFile,max_path)  = 0 then exit;
  VerInfoSize := GetFileVersionInfoSize(ForFile, Dummy);
  if VerInfoSize  = 0 then exit;
  GetMem(VerInfo, VerInfoSize);
  GetFileVersionInfo(ForFile, 0, VerInfoSize, VerInfo);
  VerQueryValue(VerInfo, '\', Pointer(VerValue), VerValueSize);
  with VerValue^ do begin
    V1 := dwFileVersionMS shr 16;
    V2 := dwFileVersionMS and $FFFF;
    V3 := dwFileVersionLS shr 16;
    V4 := dwFileVersionLS and $FFFF;
  end;
  FreeMem(VerInfo, VerInfoSize);
end;

Function _ForcebackSlash(const Dir:String):String;
begin
  Result := Dir;
  if (length(Dir)>0) and not (Dir[length(Dir)] in [#0,':','\']) then Result := Dir + '\';
end;

Procedure ParseWords(Const Sentence:string; Separators: tCharSet; List: tStringList);
var
  j      :Integer;
  inWord :boolean;
  s      : string;
begin
  List.Clear;
  inWord := false;
  s      := '';
  for j := 1 to length(Sentence) do  begin
    if Sentence[j] in Separators then begin
      if inWord then begin
        List.Add(s);
        s := '';
      end;
      inWord := false
    end else begin
      inWord := true;
      s := s + Sentence[j];
    end;
  end;
  if inWord then List.Add(s);
end;

function SetPrivDir: string;
var A: array [0..max_Path] of Char;
    L: Longint;
begin
  L := GetTempPath(max_Path,A);
  if (L = 0) OR (L > max_Path) then raise Exception.CreateRes(msg_Unabletosetprivatedirectory);
{$IFDEF WIN32}
  result := A;
{$ELSE}
  result := StrPas(A);
{$ENDIF}
  result := _forcebackslash(Result)+ChangeFileExt(ExtractFilename(Paramstr(0)),'');
  if not directoryExists(result) then Forcedirectories(result);
  session.Privatedir := result;
end;

Procedure CheckDatabaseInUse(const DBPath: string);
var s: string;
begin
  s := _ForcebackSlash(DBPath)+PdoxUsrsLck;
{ If the file exists and cannot be deleted then the databas is in use }
  if FileExists(s) and not DeleteFile(s) then
    raise EDatabaseInUse.CreateRes(msg_Databaseinuse);
end;

function GetBDEErrorMsg(rslt: DBIResult): string;
var a: array [0..DBIMAXMSGLEN] of Char;
begin
  if DbiGetErrorString (rslt,a) <> DBIERR_NONE then
    Result := ''
  else
{$IFDEF WIN32}
    Result := a
{$ELSE}
    Result := StrPas(a)
{$ENDIF}
end;

function GetTUTILITYString(ResID: integer): string;
var a: array [0..DBIMAXMSGLEN] of Char;
    numchars,h: Longint;
begin
  Result := '';
  h := GetModulehandle({$IFDEF WIN32}TU32{$ELSE}UTILITYDLL{$ENDIF});
  if h <> 0 then begin
    numchars := LoadString(h,ResID,@a,DBIMAXMSGLEN);
    if numchars = 0 then
      Result := ''
    else
  {$IFDEF WIN32}
      Result := a
  {$ELSE}
      Result := StrPas(a)
  {$ENDIF}
  end;
end;

{$IFNDEF WIN32}
function GetEnvVar(VarName: PChar): PChar;
var
  L: Word;
  P: PChar;
begin
  L := StrLen(VarName);
  P := GetDosEnvironment;
  while P^ <> #0 do
  begin
    if (StrLIComp(P, VarName, L) = 0) and (P[L] = '=') then
    begin
      GetEnvVar := P + L + 1;
      Exit;
    end;
    Inc(P, StrLen(P) + 1);
  end;
  GetEnvVar := nil;
end;

Function GetTempPath(nBufferLength: word; lpBuffer: pChar): word;
var p: pChar;
begin
  Fillchar(lpBuffer^,nBufferLength,#0);
  p := GetEnvVar('TEMP');
  if p=nil then p := GetEnvVar('TMP');
  if p=nil then result := 0 else begin
    result := StrLen(p);
    if result < nBufferLength then
      StrCopy(lpBuffer, p);
  end;
end;

function ExtractFileDrive(const FileName: string): string;
var
  I, J: Integer;
begin
  if (Length(FileName) >= 2) and (FileName[2] = ':') then
    Result := Copy(FileName, 1, 2)
  else if (Length(FileName) >= 2) and (FileName[1] = '\') and
    (FileName[2] = '\') then
  begin
    J := 0;
    I := 3;
    While (I < Length(FileName)) and (J < 2) do
    begin
      if FileName[I] = '\' then Inc(J);
      if J < 2 then Inc(I);
    end;
    if FileName[I] = '\' then Dec(I);
    Result := Copy(FileName, 1, I);
  end else Result := '';
end;

function ExtractFileDir(const FileName: string): string;
var
  I: Integer;
begin
  I := Length(FileName);
  while (I > 0) and not (FileName[I] in ['\', ':']) do Dec(I);
  if (I > 1) and (FileName[I] = '\') and
    not (FileName[I - 1] in ['\', ':']) then Dec(I);
  Result := Copy(FileName, 1, I);
end;
{$ELSE}
Procedure MyPx3PProtect(Const Px3PKey: string);
Var Rslt: Integer;
begin
  if not LoadPx3P then
    raise Exception.CreateRes(msg_cannotLoadPx3PDll);
  Rslt := Px3PProtect(pChar(Px3PKey));
  if Rslt <> 0 then
    raise Exception.CreateResFmt(msg_ErrorInPx3PProtect,[Rslt]);
end;
{$ENDIF}

end.
