{ This INC file goes into Main.PAS in both 32 bit and 16 bit Pdxrbld version }

type
{ Structure that holds command line parameters }
  tCmdParams = record
    valid       : boolean;
    IsAlias     : boolean;
    datafolder  : string;
    alias       : string;
    table       : string;
    AutoRepair  : integer;
    DoPack      : boolean;
    LogToFile   : boolean;
    QuietMode   : boolean;
  end;

Const
  LogFileExt        = '.LOG';     { log file extension (Pdxrbld.LOG in EXE directory) }
{$IFDEF WIN32}
  BDECore           = 'IDAPI32.DLL';
  UTILITY           = TU32;
{$ELSE}
  BDECore           = 'IDAPI01.DLL';
  UTILITY           = UTILITYDLL;
{$ENDIF}
  Inprise           = 'Inprise';
  InpriseTutilPage  = 'http://www.inprise.com/devsupport/bde/utilities.html';
  TblnamePadLen     = 15;      { table name pad length in the LOG file }
  TUTilEqualSection = 'EQUALS';{ INI file section holding information about TUtility version equivalence.
                                 See Readme.txt Installation notes for details. }


VAR
  LogToFile,
  Stopped: Boolean;
  LogFilepath,
  Seldatbase,
  Seltable,
  TblToList,
  LastTbl: string;
  lastTblErrCount: integer;
  LogFile: Text;
  PackOptionOn,
  IniFileRead: Boolean;
  CmdParams  : tCmdParams;
{$IFDEF WIN32}
  UseExitCode: integer = 0;
{$ENDIF}

Function CheckTUTILVersion: Boolean;
{ Compares versions of BDE core and tutil32.dll/tutility.dll. Major, minor
  and release version numbers should match. There are, however, some exceptions
  to this compatibility rule. The exceptions known by the current Pdxrbld version
  release date are incorporated into the following code. Additional exceptions
  arising in the future can be described in the INI file. If the BDE core and
  tutil32.dll/tutility.dll versions are not compatible, routine pops up
  error message and returns false. 16 bit version also checks whether
  TUTILITY is at least 2.52. }
VAR
  _V1, _V2, _V3, _V4,
   V1, V2, V3, V4: Word;
   Match: Boolean;
   SIni,STUtility,SBDECore: String;
   IniF: TRKSIniFile;
begin
  Result := false;
  _V1 := 0;
  _V2 := 0;
  _V3 := 0;
  _V4 := 0;
  GetBuildInfo(GetModuleHandle(UTILITY), V1, V2, V3, V4);
  GetBuildInfo(GetModuleHandle(BDECore), _V1, _V2, _V3, _V4);
  if _V1=0 then begin
    Showmessage(LoadStr(msg_BDENotInstalled));
  end else begin
    match := (V1=_V1) and (V2=_V2) and (V3=_V3);
    if not match
      { BDE Core version and tUtility version do not match ... }
    then begin
      { .. and this can be due to the unavailability of matching tUtility version.
        Let's check whether the current tUtility version is equivalent to the required one :}
{$IFDEF WIN32}
      {1. Hardwire the match for known compatible versions.
          TUTIL32.DLL ver 5.00 can also be used together with BDE 5.01, 5.10 and 5.11}
      if (V1=5) and (V2=0) and (V3=0) and (_V1=5) and
         ((_V2=0) and (_V3=1) or
          (_V2=1) and (_V3=0) or
          (_V2=1) and (_V3=1)) then match := True;
{$ENDIF}          
      if not match then begin
        {2. Check INI file for compatibility description
            In its EQUALS section, INI file can hold pairs  Name=Value, where "Name"
            is the BDE version string and "Value" is tUtility version string of
            TUTILITY version that should be used with that BDE instead of the default
            matching version.
            Pdxrbld versions prior to 3.30 used the version strings where numbers were
            separated by periods. Example - tutil32.dll ver 5.00 is compatible with
            BDE 5.01, which can be indicated by the following entry in the [EQUALS]
            section of the INI file:
                5.0.1=5.0.0
            This, however, was reported to cause problems on (at least some)
            WinNT boxes. Therefore Pdxrbld 3.30 and above use different version strings:
            Now the strings do not include periods and start with 'BDE'. The above shown
            entry thus should be changed to
            
                BDE501=BDE500

            Old version string format is still recognised for backward compatibility.
         }
        IniF := tRKSIniFile.Create(ChangeFileExt(paramstr(0),'.INI'));
        Try
        {   A. old INI file format }
          STUtility := Format('%d.%d.%d',[V1, V2, V3]);
          SBDECore  := Format('%d.%d.%d',[_V1, _V2, _V3]);
          SIni  := IniF.ReadString(TUTilEqualSection,SBDECore,'');
          match := SIni = STUtility;
          if not match then begin
          {   B. new INI file format }
            STUtility := Format('BDE%d%d%d',[V1, V2, V3]);
            SBDECore  := Format('BDE%d%d%d',[_V1, _V2, _V3]);
            SIni  := IniF.ReadString(TUTilEqualSection,SBDECore,'');
            match := SIni = STUtility;
          end;
        Finally
          IniF.Free;
        End;
      End;
    End;
    Result := match;
    if not Result then
      Showmessage(Format(LoadStr(msg_TUTILVersionmismatch),[UTILITY,_V1,_V2,_V3,Inprise,InpriseTutilPage]))
{$IFNDEF WIN32}
   { TUTILITY 2.52 or higher required }
    else if (V1 < 2) or ((V1 = 2) and
                    ((V2 < 5) or (V2 = 5) and
                     ((V3 < 1) or (V3 = 1) and
                      (V4 < 52) { TUTILITY 2.52 }
                     )
                    )
                   )
    then begin
      Result := false;
      ShowMessage(LoadStr(msg_TUTILITY252Requred));
    end;
{$ENDIF}
  end;
end;

{$IFDEF WIN32}
{ Registering/unregistering application under HKLM\Software\Microsoft\Windows\CurrentVersion\App Paths
  key so it does not require path in the environment - now you can ShellExecute it without knowing
  its exact location. Called from within main form OnCreate with DoRegister=true. }
procedure RegisterApp(DoRegister: boolean); { True = register, false=unregister }
{$IFDEF VER90}
    procedure makesinglebackslashes(var s: string);
    var i: integer;
    begin
      repeat
        i := pos('\\',s);
        if i > 0 then System.Delete(s,i,1);
      until i=0
    end;
{$ENDIF}
var REGSTRPATHAPPPATHS: String;
begin
  REGSTRPATHAPPPATHS:=REGSTR_PATH_APPPATHS;
{$IFDEF VER90} // D2 key constanst have double backslashes
  makesinglebackslashes(REGSTRPATHAPPPATHS);
{$ENDIF}
  with tRegistry.Create do Try
    RootKey := HKEY_LOCAL_MACHINE;
    if DoRegister then begin
      if OpenKey(REGSTRPATHAPPPATHS+'\'+ExtractFilename(ParamStr(0)),True) then begin
        WriteString('',ParamStr(0));
        WriteString('Path',ExtractFileDir(ParamStr(0)));
      end
    end else begin
      if OpenKey(REGSTRPATHAPPPATHS,False) then
        DeleteKey(ExtractFilename(ParamStr(0)));
    end;
  finally
    Free;
  end;
end;
{$ENDIF}

Procedure Initialize;
{
  In 32 bits checks for  registration/unregistration parameters in the command line.
  If these parameters are present, performs registration/unregistration and
  quits immediately.
  Also checks whether BDE is installed (32 bit only) and whether BDE and TUTILITY
  versions do match.
  Called from within initialization section.
}
begin
{$IFDEF WIN32}
  if paramcount=1 then begin
    if CompareText(paramstr(1),'REGISTER')=0 then begin
      RegisterApp(True);
      halt;
    end else
    if CompareText(paramstr(1),'UNREGISTER')=0 then begin
      RegisterApp(False);
      halt;
    end;
  end;
  Try
    session.active := true;
  Except
    On E: Exception do begin
      Showmessage(E.message);
      UseexitCode := -1;
      halt;
    end;
  End;
  if not CheckTUTILVersion then begin
    UseexitCode := -1;
    halt;
  end;
{$ELSE}
  if not CheckTUTILVersion then begin
    halt;
  end;
{$ENDIF}
end;

{ ------- CALLBACKS -------- }

{ Called from within tUtil32.dll code  }
function TableCallBack(ecbType: CBType; Data: LongInt; pcbInfo: pointer):
  CBRType;
  {$IFDEF WIN32} stdcall; {$ELSE} export;{$ENDIF}

  function GetIntegerFromPackMessage(const s: string): integer;
  { During pack (dbiDoRestructure routine), messages of the form
    'Records appended : 251' or
    'Reading record : 166' etc
    are stored in the CBPROGRESSDesc.szMsg field of the record passed to
    callback routine. GetIntegerFromPackMessage function will try to
    retrieve the number from szMsg field }
  Var NumS: string;
      i: integer;
  begin
    Result := - 1;
    NumS := '';
    For i := Length(s) downto 1 do
     if s[i] in ['0'..'9'] then NumS := s[i] + NumS
      else break;
    if NumS <> '' then Result := StrToInt(NumS);
  end;

var pcDone,IxNo: integer;
begin
  if Stopped then Result := cbrAbort
  else  Result := cbrUSEDEF;
  if pCbInfo = NIL then exit;
  pcDone := CBPROGRESSDesc(pcbInfo^).iPercentdone;
 { now set mainform.labeldetails.caption }
  case data of
  clbdataTUTILITYVerify: { callback from within  TUTILITY - verify }
                   with TUVerifyCallBack(pcbInfo^) do begin
                   {$IFDEF WIN32}
                     IxNo := ShortInt(Hi(TotalIndex))+1;  {! don't trust structure declared in TUtil32.pas }
                   {$ELSE}
                     IxNo := ShortInt(Hi(CurrentIndex))+1;
                   {$ENDIF}
                     case process of
                     TUVerifyData       : mainform.labelDetails.Caption := loadStr(msg_Tabledata);
                     TUVerifySXIntegrity: mainform.labelDetails.Caption := Format(loadStr(msg_IndexNo),[IxNo]);
                     else pcDone := - 1;
                     end;
                   end;
  clbdataTUTILITYRebuild: { callback from within  TUTILITY - verify }
                   with TUVerifyCallBack(pcbInfo^) do begin
                     mainform.labelDetails.caption := '';
                   end;
  clbdataPDXRBLD : { callback from within Rbld1 unit }
                   with CBPROGRESSDesc(pcbInfo^) do begin
                     { nothing }
                   end;
  clbDataPack    : { callback from within  BDE (pack process) }
                   { You may need to translate string literals if localized
                     version of BDE is used }
                   with tCBRESTRPROGRESSDesc(pcbInfo^) do begin
                     if StrPos(szMsg,'Reading') <> nil then begin
                       mainform.labelDetails.Caption := loadStr(msg_ReadingRecords);
                     end else
                     if StrPos(szMsg,'Writing') <> nil then begin
                       mainform.labelDetails.Caption := loadStr(msg_WritingRecords);
                     end else
                     if StrPos(szMsg,'Sorting') <> nil then begin
                       mainform.labelDetails.Caption := loadStr(msg_SortingRecords);
                     end else begin  { 'Records appended ' }
                       pcDone := GetIntegerFromPackMessage(szMsg);
                       if pcDone >= 0 then begin
                         mainform.labelDetails.Caption := loadStr(msg_AppendingRecords);
                         if TotalRecords > 0 then begin
                           pcDone := round(100*pcDone/TotalRecords);
                         end;
                       end
                     end;
                   end;
  end;
  if pcDone >= 0 then
    mainform.Progresstable.{$IFDEF WIN32}Position{$ELSE}Progress{$ENDIF} := pcDone
  else
    mainform.Progresstable.{$IFDEF WIN32}Position{$ELSE}Progress{$ENDIF} := 0;
  Application.ProcessMessages;
end;


procedure  LogTableIsOK;
{ Write "table is OK" to Log file ..}
begin
  if LogToFile             and  { Write to log }
     (LastTbl  <> '')      and  { table processing finished }
     (LastTblErrCount = 0) { there were no errors } then  begin
     Writeln(LogFile,LoadStr(msg_TblnamePrefix),
                     Padright(LastTbl,TblnamePadLen),
                     LoadStr(msg_MsgSeparator),
                     LoadStr(msg_noerrorsfound));
  end;
end;

{ Called from within CheckAndRepairTables routine }
function DatabaseCallback(ecbType: CBType; Data: LongInt; pcbInfo: PCallbackRecord): CBRType;
   {$IFDEF WIN32} stdcall; {$ELSE} far; {$ENDIF}
begin
  with mainform do begin
    ProgressDatabase.{$IFDEF WIN32}Position{$ELSE}Progress{$ENDIF} := pCBInfo^.percentdone;
    if LastTbl <> Format('%s%s',[pCBInfo^.tablename,pCBInfo^.extension]) then begin
      if not CheckDoPack.Checked then LogTableIsOK; { Write "table is OK" to logFile if lastTblErrCount=0 }
                                                    { if CheckDoPack.Checked=true then "table is OK" is
                                                      written before actual packing }
      labelWhatDoing.caption := '';
      LastTbl := Format('%s%s',[pCBInfo^.tablename,pCBInfo^.extension]);
      lastTblErrCount := 0;
      TblToList := LastTbl;
      LabelTablename.Caption := LastTbl;
    end;
    case pCBInfo^.aProcess of
    cpVerify: begin
                case pCBInfo^.aState of
                cstStarting       : begin
                                      labelWhatDoing.caption := LoadStr(msg_Checkingforerrors);
                                    end;
                cstReportingErrors: if pCBInfo^.Errormessage <> '' then
                                        AddListViewMessage(TblToList,pCBInfo^.Errormessage,pCBInfo^.aProcess);
                cstReportingResult: begin
                                      if pCBInfo^.ErrorCode <> 0 then
                                        AddListViewMessage(TblToList,pCBInfo^.Errormessage,pCBInfo^.aProcess);
                                    end;
                end
              end;
    cpRepair: begin
                case pCBInfo^.aState of
                cstStarting       : begin
                                      labelWhatDoing.caption := LoadStr(msg_Repairing);
                                    end;
                cstReportingResult: begin
                                      if pCBInfo^.Errormessage <> '' then
                                        AddListViewMessage(TblToList,pCBInfo^.Errormessage,pCBInfo^.aProcess);
                                    end;
                end
              end;
    cpOrphansCheck: begin
                case pCBInfo^.aState of
                cstStarting       : begin
                                      labelWhatDoing.caption := LoadStr(msg_CheckingOrphans);
                                    end;
                cstReportingErrors: if pCBInfo^.Errormessage <> '' then
                                        AddListViewMessage(TblToList,pCBInfo^.Errormessage,pCBInfo^.aProcess);
                cstReportingResult: begin
                                      if pCBInfo^.ErrorCode <> 0 then
                                        AddListViewMessage(TblToList,pCBInfo^.Errormessage,pCBInfo^.aProcess);
                                    end;
                end
              end;
    cpPack  : begin
                case pCBInfo^.aState of
                cstStarting       : begin
                                      labelWhatDoing.caption := LoadStr(msg_Packing);
                                      mainform.labelDetails.Caption := loadStr(msg_AppendingRecords);
                                      mainform.Progresstable.
                                        {$IFDEF WIN32}Position{$ELSE}Progress{$ENDIF} := 0;
                                      LogTableIsOK;
                                    end;
                cstReportingResult: begin
                                      if pCBInfo^.ErrorCode <> 0 then
                                         AddListViewMessage(TblToList,pCBInfo^.Errormessage,pCBInfo^.aProcess)
                                      else
                                      if LogToFile then
                                         Writeln(LogFile,
                                                 LoadStr(msg_TblnamePrefix),
                                                 padright(LastTbl,TblnamePadLen),
                                                 LoadStr(msg_MsgSeparator),
                                                 LoadStr(msg_successfulypacked));
                                    end;
                end;
              end
    end;
  end;
  application.processmessages;
  if Stopped then Result := cbrAbort
  else  Result := cbrUSEDEF;
end;

function SelectDBPath(Const CorruptDB,CorruptionDescr: String): String;
{$IFNDEF WIN32}
far;
{$ENDIF}
{ SELECTS DATABASE FOR HEADER BORROW. Passed as procedural
  parameter to CheckAndRepairTables routine. Returns user
  selected path to header source backup tables.
}
var
  s: string;
begin
  Result := '';
  case MessageDlg(Format(LoadStr(msg_DoYouWantToBorrowHeader),
                  [ChangeFileExt(CorruptDB,''),
                   CorruptionDescr]),mtConfirmation,[mbYes,mbNo,mbAbort],0) of
  mrYes: begin
           s := '';
          {$IFDEF WIN32}
           if SelectDirectory(LoadStr(msg_Selectdatabasefolder),s,'',s) then
          {$ELSE}
           if SelectDirectory(s,[],0) then
          {$ENDIF}
             Result := s;
         end;
  mrNo: ;
  mrAbort: Sysutils.Abort;
  end;
end;


Procedure  ParseCmdLine;
{ parses command line and fills Cmdparams record.

  Command line syntax:

Pdxrbld </F<Folder_path>|/A<Alias>> [/T<TableName>] [[/X<Px3P key>]...] [[/P<password>]...]
                                         [-R+|-R-|-R0|-R1|-R2] [-P+|-P-] [-L+|-L-] [-Q+|-Q-]

where

either Folder_path (/F parameter) or Alias (/A parameter) must specify database location

/T parameter (optional) specify table which should be checked/repaired. Default= All tables
/X parameter (optional-32 bits only) specify Px3P key. All password (/P) parameters occurring
   *after* this parameter are considered associated with this Px3P key.
   This parameter can occur several times followed by the corresponding passwords.
   Example:
       Pdxrbld /ADBDEMOS /PJohn /Xfoo /PBrian /Xaargh /PJenny
       here password "John" is associated with default encryption,
            password "Brian" is associated with Px3P encryption with key "foo" and
            password "Jenny" is associated with Px3P encryption with key "aargh"

/P parameter (optional) specify master password. This parameter can occur several times.


-R switch sets "rebuild scope" option. Default = 0 (Don't rebuild)
   +|- switch options are still recognised to maintain the compatibility with previous version
   -R- sets rebuild scope to "Don't rebuild"
   -R+ sets rebuild scope to "Rebuild corrupted tables"
   in the current version, you should use number after -R switch:
   -R0 sets rebuild scope to "Don't rebuild"
   -R1 sets rebuild scope to "Rebuild corrupted tables"
   -R2 sets rebuild scope to "Rebuild all"
-P switch sets "pack tables" option On/Off. Default = Off
-L switch sets "log to file" option On/Off. Default = Off
-Q switch sets "quiet mode" option On/Off. Default = Off
   In quiet mode, application shuts down after performing the check/rebuild
   without popping up the messagebox that shows the check/rebuild result.
}
var i: integer;
    s: string;
{$IFDEF WIN32}
    CurrPx3PKey: String;
{$ENDIF}
begin
{$IFDEF WIN32}
  CurrPx3PKey := '';
{$ENDIF}
  CmdParams.Valid := False;
  if paramcount > 0 then begin
    CmdParams.Valid := True;
    for i := 1 to paramCount do begin
      s := AnsiUppercase(paramstr(i));
      if pos('/F',s) = 1 then begin { databasepath }
        CmdParams.IsAlias := false;
        CmdParams.datafolder := copy(paramstr(i),3,length(s)-2);
      end else
      if pos('/A',s) = 1 then begin { alias }
        CmdParams.IsAlias := True;
        CmdParams.Alias := copy(paramstr(i),3,length(s)-2);
      end else
      if pos('/T',s) = 1 then begin { table }
        CmdParams.table := copy(paramstr(i),3,length(s)-2);
        if CmdParams.table <> '' then
          CmdParams.table := ChangeFileExt(CmdParams.table,'.DB');
      end else
{$IFDEF WIN32}
      if pos('/X',s) = 1 then begin { Px3P key }
        CurrPx3PKey := copy(paramstr(i),3,length(s)-2);
        AddSessionPassword(CurrPx3PKey,'');
      end else
{$ENDIF}
      if (pos('/P',s) = 1) and (Length(s)>=3) then begin { password }
{$IFNDEF WIN32}
        AddSessionPassword(copy(paramstr(i),3,length(s)-2));
{$ELSE}
        AddSessionPassword(CurrPx3PKey,copy(paramstr(i),3,length(s)-2));
{$ENDIF}
      end else
      if (pos('-R',s) = 1) and (Length(s)=3) then begin { rebuild scope switch }
        if (s[3] = '+') then
          CmdParams.AutoRepair := rsCorrupted
        else
        if (s[3] = '0') then
          CmdParams.AutoRepair := rsNone
        else
        if (s[3] = '1') then
          CmdParams.AutoRepair := rsCorrupted
        else
        if (s[3] = '2') then
          CmdParams.AutoRepair := rsAll
        else
        if (s[3] = '3') then
          CmdParams.AutoRepair := rsAllSkipTUVerify
        else
          CmdParams.AutoRepair := rsNone;
      end else
      if (pos('-P',s) = 1) and (Length(s)=3) and (s[3] in ['+','-']) then begin { pack switch }
        CmdParams.DoPack := s[3] = '+';
      end else
      if (pos('-L',s) = 1) and (Length(s)=3) and (s[3] in ['+','-']) then begin { log to file switch }
        CmdParams.LogToFile := s[3] = '+';
      end else
      if (pos('-Q',s) = 1) and (Length(s)=3) and (s[3] in ['+','-']) then begin { quiet mode switch }
        CmdParams.QuietMode := s[3] = '+';
      end
    end
  end
end;


function  TMainFormGetAfterprocessInfo: string;
{ Gets the check/repair result description }
begin
{$IFDEF WIN32}
  if mainform.ListView1.Items.Count > 0 then
{$ELSE}
  if mainform.msgMessages.Count > 0 then
{$ENDIF}
    Result := LoadStr(msg_Therewereerrors)
  else if not PackOptionOn then
    Result := LoadStr(msg_TablesappeartobeOK)
  else begin
    Result := LoadStr(msg_AlltablesSuccesfulyPacked)
  end;
end;

procedure TMainFormButtonBrowseClick(Sender: TObject);
{ Brings up "Browse for folder" dialog }
var s: string;
begin
  s := mainform.DirectoryCombo.Text;
{$IFDEF WIN32}
  if SelectDirectory(LoadStr(msg_Selectdatabasefolder),s,'',s) then begin
{$ELSE}
  if SelectDirectory(s,[],0) then begin
{$ENDIF}
    if mainform.DirectoryCombo.Text <> s then mainform.tableCombo.Items.Clear;
    mainform.DirectoryCombo.Text := s
  end;
end;

procedure TMainFormUpdateradio;
{ Enables/Disables controls according to radioAlias selection }
begin
  case mainform.RadioAlias.ItemIndex of
  0: begin
       mainform.DirectoryCombo.Enabled := false;
       mainform.ButtonBrowse.Enabled := false;
       mainform.AliasCombo.Enabled := True;
    end;
  1: begin
       mainform.DirectoryCombo.Enabled := True;
       mainform.ButtonBrowse.Enabled := True;
       mainform.AliasCombo.Enabled := False;
     end;
  end;
end;

procedure TMainFormRadioAliasClick(Sender: TObject);
begin
  TMainFormUpdateradio;
  mainForm.tableCombo.Items.Clear;
end;

procedure TMainFormOpenDataBaseList;
{ Gets the list of available aliases }
var
  TmpCursor: hDbiCur;
  vDBDesc: DBDesc;
begin
  mainform.AliasCombo.Items.Clear;
  Check(DbiOpenDatabaseList(TmpCursor));
  while (DbiGetNextRecord(TmpCursor, dbiNOLOCK, @vDBDesc, nil)
                                      = DBIERR_NONE) do begin
    if StrIComp(vDBDesc.szDBType,'STANDARD')=0 then
      mainform.AliasCombo.Items.Add(vDBDesc.szName);
  end;
  Check(DbiCloseCursor(TmpCursor));
  if mainform.AliasCombo.Items.Count > 0 then begin
    mainform.AliasCombo.ItemIndex := 0
  end else begin
    mainform.AliasCombo.Enabled := false;
    mainform.RadioAlias.ItemIndex := 1;
    mainform.RadioAlias.Enabled := false;
  end;
end;

procedure TMainFormSetTexts;
var   verstr: string;
      V1, V2, V3, V4: Word;
begin
  GetBuildInfo(hInstance, V1, V2, V3, V4);
  VerStr  := format('%d.%d%d',[V1,V2,V3]);
  mainform.Caption := LoadStr(dfm_MainFormCaption)+VerStr
  {$ifndef WIN32}+' (16 bit)'{$endif};
{$IFDEF WIN32}
  mainform.TabSheet1.Caption := LoadStr(dfm_MainFormTabSheet1Caption);
  mainform.TabSheet2.Caption := LoadStr(dfm_MainFormTabSheet2Caption);
{$ELSE}
  mainform.PageControl1.Pages[0] := LoadStr(dfm_MainFormTabSheet1Caption);
  mainform.PageControl1.Pages[1] := LoadStr(dfm_MainFormTabSheet2Caption);
{$ENDIF}
  mainform.ButtonStart.Caption := LoadStr(dfm_MainFormButtonStartCaption);
  mainform.ButtonStop.Caption := LoadStr(dfm_MainFormButtonStopCaption);
  mainform.ButtonClose.Caption := LoadStr(dfm_MainFormButtonCloseCaption);
  mainform.Label1.Caption := LoadStr(dfm_MainFormLabel1Caption);
  mainform.Label2.Caption := LoadStr(dfm_MainFormLabel2Caption);
  mainform.Label3.Caption := LoadStr(dfm_MainFormLabel3Caption);
  mainform.ButtonBrowse.Caption := LoadStr(dfm_MainFormButtonBrowseCaption);
  with mainform.rgRebuildScope do begin
    Caption := LoadStr(dfm_MainFormrgRebuildScopeCaption);
    Items.Clear;
    Items.Add(LoadStr(dfm_MainFormCheckDontRepairCaption));
    Items.Add(LoadStr(dfm_MainFormCheckAutoRepairCaption));
    Items.Add(LoadStr(dfm_MainFormCheckAllRepairCaption));
  end;
  mainform.CheckDoPack.Caption := LoadStr(dfm_MainFormCheckDoPackCaption);
  mainform.CheckDoLog.Caption := LoadStr(dfm_MainFormCheckDoLogCaption);
  mainform.CheckSkipVerify.Caption := LoadStr(dfm_MainFormCheckSkipVerifyCaption);
  mainform.ButtonViewLog.Caption := LoadStr(dfm_MainFormButtonViewLogCaption);
  With mainform.radioAlias do begin
    Caption := LoadStr(dfm_MainFormRadioAliasCaption);
    Items.Clear;
    Items.Add(LoadStr(dfm_RadioAliasByAlias));
    Items.Add(LoadStr(dfm_RadioAliasByDirectory));
  end;
end;

procedure UpdateRepairScopeControls(RepairScope: Integer);
begin
  if RepairScope < rsAllSkipTUVerify then begin
    mainform.rgRebuildScope.ItemIndex := RepairScope;
    mainform.CheckSkipverify.Checked := false;
  end else begin
    mainform.rgRebuildScope.ItemIndex := rsAll;
    mainform.CheckSkipverify.Checked := True;
  end;
  mainform.CheckSkipverify.Enabled := mainform.rgRebuildScope.ItemIndex = rsAll;
end;

Function GetRepairScope: Integer;
begin
  Result := mainform.rgRebuildScope.ItemIndex;
  if (Result = rsAll) and mainform.CheckSkipverify.Checked then Result := rsAllSkipTUVerify;
end;

procedure TMainFormFormCreate(Sender: TObject);
var i: integer;
    s: string;
begin
  {$IFDEF WIN32}
  RegisterApp(True);
  {$ENDIF}
  tMainFormsettexts;
  mainform.Labeltablename.visible := False;
  with mainform.pagecontrol1 do begin
    activepage := pages[0];
  {$IFDEF WIN32}
    pages[1].tabVisible := false;
  {$ELSE}
    controls[1].Visible := false;
  {$ENDIF}
  end;
  mainform.Image1.Width  := mainform.Bevel1.Width - 2;
  mainform.Image1.Height := mainform.Bevel1.Height - 2;
  mainform.Image1.Top    := mainform.Bevel1.Top + 1;
  mainform.Image1.Left   := mainform.Bevel1.Left + 1;
  TMainFormOpenDataBaseList;
 { Reads settings stored in INI file: }
  With tRKSIniFile.Create(ChangeFileExt(paramstr(0),'.INI')) do Try
    i := 1;
    While true do begin
      s := ReadString('PATHS','P'+IntToStr(i),'');
      if s='' then break else mainform.DirectoryCombo.Items.Add(s);
      inc(i);
    end;
    mainform.DirectoryCombo.Text := ReadString('PATHS','PATH','');
    s := ReadString('PATHS','ALIAS','');
    if (s<>'') and (mainform.AliasCombo.Items.IndexOf(s) >= 0) then
      mainform.AliasCombo.ItemIndex := mainform.AliasCombo.Items.IndexOf(s);
    mainform.CheckSkipverify.Checked := False;
    i := ReadInteger('OPTIONS','SCREPAIR',-1);
    if not (i in [rsNone..rsAllSkipTUVerify]) then begin
      if ReadBool('OPTIONS','REPAIR',false) then
        i := rsCorrupted
      else
        i := rsNone;
    end;
    UpdateRepairScopeControls(I);
    mainform.CheckDoPack.Checked := ReadBool('OPTIONS','PACK',false);
    mainform.CheckDoLog.Checked := ReadBool('OPTIONS','LOG',false);
    if mainform.RadioAlias.Enabled then
      mainform.RadioAlias.ItemIndex := ReadInteger('OPTIONS','SELECT',0);
  Finally
    Free;
  End;
  TmainformUpdateradio;
  IniFileRead := True;
  mainform.Labeltablename.caption := '';
  mainform.LabelDetails.caption := '';
  mainform.LabelWhatDoing.caption := '';
  mainform.labelInfo.caption := LoadStr(msg_Ready);
  LogFilepath := ChangeFileExt(paramstr(0),LogFileExt);
  mainform.ButtonViewLog.Visible := FileExists(LogFilepath);
  ParseCmdLine;
  if CmdParams.valid then PostMessage(mainform.handle,wm_runcmdline,0,0);
end;

procedure TMainFormFormDestroy(Sender: TObject);
var i: integer;
begin
 { Stores current settings to INI file: }
  if IniFileRead then
  With tRKSIniFile.Create(ChangeFileExt(paramstr(0),'.INI')) do Try
    EraseSection('PATHS');
    For i := 1 to mainform.DirectoryCombo.Items.Count do
      WriteString('PATHS','P'+IntToStr(i),mainform.DirectoryCombo.Items[i-1]);
    WriteString('PATHS','PATH',mainform.DirectoryCombo.Text);
    if mainform.AliasCombo.Text<>'' then WriteString('PATHS','ALIAS',mainform.AliasCombo.Text);
    WriteInteger('OPTIONS','SCREPAIR',GetRepairScope);
    WriteBool('OPTIONS','PACK',mainform.CheckDoPack.Checked );
    WriteBool('OPTIONS','LOG',mainform.CheckDoLog.Checked );
    WriteInteger('OPTIONS','SELECT',mainform.RadioAlias.ItemIndex);
  Finally
    Free;
  End;
end;

procedure InitLogFile(DoLog: Boolean);
{ Initializes LOG file }
begin
  LogToFile := false;
  if DoLog then begin
    assignFile(LogFile,LogFilepath);
    Try
      if FileExists(LogFilepath) then begin
        Append(LogFile);
        Writeln(LogFile,'--------');
      end else
        Rewrite(LogFile);
      LogToFile := True;
    Except
      On E: Exception do begin
        if MessageDlg(Format(LoadStr(msg_CannotcreateopenLOGfile),[E.message]),mtWarning,[mbYes,mbNo],0) = mrNo then
          sysutils.Abort;
      end;
    End;
  end;
end;

procedure TMainFormDoneLogFile;
{ Closes log file }
begin
  if LogToFile then CloseFile(LogFile);
  mainform.ButtonViewLog.Visible := FileExists(LogFilepath);
end;

procedure TMainFormInitprocess;
{ Initializes controls and variables before check/repair process }
begin
  Stopped   := false;
  TblToList := '';
  LastTbl   := '';
{$IFDEF WIN32}
  mainform.Pagecontrol1.pages[1].tabVisible := false;
{$ELSE}
  mainform.Pagecontrol1.controls[1].Visible := false;
  mainform.msgMessages.Clear;
{$ENDIF}
  mainform.ListView1.Items.Clear;
  mainform.Labeltablename.caption := '';
  mainform.LabelWhatDoing.caption := '';
  mainform.LabelDetails.caption := '';
  mainform.labelInfo.caption := LoadStr(msg_Nowprocessing);
  mainform.Progressdatabase.{$IFDEF WIN32}Position{$ELSE}progress{$ENDIF}:= 0;
  mainform.Progresstable.{$IFDEF WIN32}Position{$ELSE}progress{$ENDIF} := 0;
  mainform.Panel1.Enabled := false;
  mainform.ButtonViewLog.Enabled := false;
  mainform.ButtonStart.Enabled := false;
  mainform.ButtonClose.Enabled := false;
  mainform.ButtonStop.Enabled := True;
  if LogToFile then begin
    Writeln(LogFile,LoadStr(msg_Datetime),dateTimeToStr(now));
    Writeln(LogFile,LoadStr(msg_Database),Seldatbase);
    if Seltable <> '' then
    Writeln(LogFile,LoadStr(msg_Selectedtable),Seltable);
    Writeln(LogFile);
  end;
  application.processmessages;
  mainform.Labeltablename.visible := true;
  Screen.cursor := crHourglass;
end;

procedure TMainFormDoneprocess;
{ sets control values after check/repair process finished }
begin
  Screen.cursor := crDefault;
  if not PackOptionOn then LogTableIsOK;
  mainform.Labeltablename.caption := '';
  mainform.LabelDetails.caption := '';
  mainform.LabelWhatDoing.caption := '';
  mainform.labelInfo.caption      := '';
  mainform.Progressdatabase.{$IFDEF WIN32}Position{$ELSE}progress{$ENDIF} := 0;
  mainform.Progresstable.{$IFDEF WIN32}Position{$ELSE}progress{$ENDIF} := 0;
  mainform.Panel1.Enabled := True;
  mainform.ButtonViewLog.Enabled := True;
  mainform.ButtonStart.Enabled := True;
  mainform.ButtonClose.Enabled := True;
  mainform.ButtonStop.Enabled := False;
  mainform.Labeltablename.visible := False;
  application.processmessages;
end;

Function TMainFormGetSelecteddatabase(Var IsAlias: boolean): String;
{ Gets databasename string from controls: }
begin
  IsAlias := mainform.RadioAlias.ItemIndex = 0;
  if IsAlias then Result := mainform.AliasCombo.Text
  else Result := mainform.DirectoryCombo.Text;
  if not (IsAlias or DirectoryExists(Result)) then Result := '';
  if Result='' then raise Exception.CreateRes(msg_Youmustenteravaliddatabase);
end;

procedure TMainFormUpdateTablecombo(Const datBname: string);
{ Updates list of tables available in the selected database }
var l: tStringList;
begin
  mainform.tablecombo.items.Clear;
  l:= tStringList.Create;
  Try
    Session.GetTableNames(datBname, '*.DB', True {Extensions}, False {SystemTables}, l);
    if l.Count > 0 then begin
      l.Sorted := true;
      mainform.tablecombo.items.Assign(l);
      mainform.tablecombo.items.Insert(0,LoadStr(msg_Alltables));
      mainform.tablecombo.itemindex := 0;
    end;
  Finally
    l.Free;
  End;
end;

procedure TMainFormButtonStartClick(Sender: TObject);
var
  scopeRepair: Integer;
  IsAlias: Boolean;
begin
  {$IFDEF WIN32}
  UseExitCode := 0;
  {$ENDIF}
  Seldatbase := tmainformGetSelecteddatabase(IsAlias);
  Seltable   := mainform.tableCombo.text;
  if Seltable=LoadStr(msg_Alltables) then Seltable := '';
  if not IsAlias then with mainform.DirectoryCombo do
    if Items.IndexOf(Seldatbase) < 0 then Items.Add(Seldatbase);
  PackOptionOn := mainform.CheckDoPack.Checked;
  scopeRepair:= GetRepairScope;
  InitLogFile(mainform.CheckDoLog.Checked);
  Try
    tmainformInitProcess;
    Try
      Try
        CheckAndRepairTables(Seldatbase,
                             IsAlias,
                             scopeRepair,
                             PackOptionOn,
                             @TableCallBack,
                             @DatabaseCallback,
                             Seltable,
                             SelectDBPath);
      Finally
        tmainformDoneProcess;
      End;
    Except
      On E: EAbort do begin
         mainform.labelInfo.caption := LoadStr(msg_Processinterruptedbyuser);
         if LogToFile then begin
           Writeln(LogFile,mainform.labelInfo.caption);
         end;
         raise
      end;
      On E: Exception do begin { Corrected in ver 3.31 - Log error and set
                                 exit code on all exceptions except user abort }
       {$IFDEF WIN32}
        if E is EDatabaseInUse then
          UseExitCode := -2
        else
          UseExitCode := -1;
       {$ENDIF}
        if LogToFile then Writeln(LogFile,E.message);
        raise
      end;
    End;
    mainform.labelInfo.caption := tmainformGetAfterprocessInfo;
  Finally
    tmainformDoneLogFile;
  End;
end;


procedure TMainFormWMRunCmdLine(Var Message:tMessage);
{ run with command line }
var
  I : Integer;
begin
  if not Cmdparams.valid then exit;
  {$IFDEF WIN32}
  UseExitCode := 0;
  {$ENDIF}
  if cmdparams.IsAlias then begin
    SelDatbase := Cmdparams.Alias;
    I := mainform.AliasCombo.Items.IndexOf(SelDatbase);
    if I < 0 then Cmdparams.valid := false
    else begin
      mainform.radioAlias.ItemIndex := 0;
      tmainformUpdateradio;
      mainform.AliasCombo.ItemIndex := i;
    end;
  end else begin
    SelDatbase := Cmdparams.datafolder;
    Cmdparams.valid := (SelDatbase<>'') and DirectoryExists(SelDatbase);
    if Cmdparams.valid then begin
      mainform.radioAlias.ItemIndex := 1;
      tmainformUpdateradio;
      mainform.DirectoryCombo.Text := SelDatbase;
    end;
  end;
  if not Cmdparams.valid then raise Exception.CreateRes(msg_Youmustenteravaliddatabase);
  tmainformUpdateTablecombo(SelDatbase);
  Seltable   := Cmdparams.table;
  I := mainform.tableCombo.Items.IndexOf(Seltable);
  mainform.tableCombo.ItemIndex := I;
  PackOptionOn  := Cmdparams.DoPack;
  UpdateRepairScopeControls(Cmdparams.AutoRepair);
  mainform.CheckDoPack.Checked := PackOptionOn;
  mainform.CheckDoLog.Checked  := Cmdparams.LogToFile;
  InitLogFile(Cmdparams.LogToFile);
  Try
    TmainformInitProcess;
    Try
      Try
        CheckAndRepairTables(Seldatbase,
                             Cmdparams.IsAlias,
                             Cmdparams.AutoRepair,
                             PackOptionOn,
                             @TableCallBack,
                             @DatabaseCallback,
                             Seltable,
                             SelectDBPath);
      Finally
        tmainformDoneProcess;
      End;
    Except
      On E: EAbort do begin
         mainform.labelInfo.caption := LoadStr(msg_Processinterruptedbyuser);
         if LogToFile then begin
           Writeln(LogFile,mainform.labelInfo.caption);
         end;
         raise { in quiet mode, application will not terminate automatically }
      end;
      On E: Exception do begin { Corrected in ver 3.31 - handle all exceptions except user abort in quiet mode }
        {$IFDEF WIN32}
        if E is EDatabaseInUse then
          UseExitCode := -2
        else
          UseExitCode := -1;
        {$ENDIF}
        if LogToFile then Writeln(LogFile,E.message);
        if not CmdParams.QuietMode then raise
      end;
    End;
    mainform.labelInfo.caption := tmainformGetAfterprocessInfo;
  Finally
    tMainFormDoneLogFile;
  End;
  if not CmdParams.QuietMode then
    showmessage(mainform.labelInfo.caption);
{$IFDEF WIN32}
  with mainform.pagecontrol1 do if pages[1].tabVisible
{$ELSE}
  with mainform.pagecontrol1 do if controls[1].Visible
{$ENDIF}
  and not CmdParams.QuietMode
  then begin
    activepage := pages[1];
    Cmdparams.valid := false;
  end else begin
    application.terminate;
  end;
end;

procedure TMainFormButtonStopClick(Sender: TObject);
begin
  Stopped := True;
end;

procedure TMainFormButtonCloseClick(Sender: TObject);
begin
  application.terminate
end;


procedure TMainFormtableComboEnter(Sender: TObject);
Var IsAlias: boolean;
begin
  TmainFormUpdateTablecombo(TMainFormGetSelecteddatabase(IsAlias));
end;

procedure TMainFormAliasComboChange(Sender: TObject);
begin
  mainform.tablecombo.items.clear;
end;

procedure TMainFormDirectoryComboChange(Sender: TObject);
begin
  mainform.tablecombo.items.clear;
end;

procedure TMainFormButtonViewLogClick(Sender: TObject);
var pszFile: array [0..max_path] of Char;
begin
  ShellExecute(mainform.handle,
               NIL,
               'Notepad.exe',
               StrPCopy(pszFile,LogFilePath),
               NIL,
               SW_SHOWNORMAL);
end;


procedure TMainFormrgRebuildScopeClick(Sender: TObject);
begin
  UpdateRepairScopeControls(GetRepairScope);
end;



