unit PxHeader;
{=====================================================================
 Pdxrbld - Paradox tables repair utility  Copyright 1998 by
 Roman Krejci-RKS, Pujmanove 1552, 140 00 Prague 4, Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz
 =====================================================================

   This unit contains description of the Paradox table header structure
   together with a class that does read and update this structure.

   based on work compiled by Randy Beck (last update 5-3-1998)
   RandyBeck@bitsmart.com   (bex@compuserve.com)

   more comprehensive summary can be found at
   http://ourworld.compuserve.com/homepages/bex/pxspecs.htm

   ... all comments by Randy Beck, except where explicitly stated
 -------------------------------------------------------------------------}

interface
uses sysutils,classes;


type
    int16 = smallint;        {2 byte signed}
    int32 = longint;         {4 byte signed}
    ptr32 = longint;         {4 byte offset}

  { pChar is 4 byte pointer to 1 byte character }
  { Pointer is 4 byte pointer to something      }
{Comments above by RK}

    TPxFieldInfo = packed record
                  fType   : byte;
                  fSize   : byte;
{+------+---------------------------------------------------------------------+
|      |        These are the field identifiers for each field in the        |
|      |        table:                                                       |
|      |                                                                     |
|      |           fType  fSize(decimal)                                     |
|      |           -------------------------                                 |
|      |            $01     v   "A"  Alpha                                   |
|      |            $02     4   "D"  Date                                    |
|      |            $03     2   "S"  Short integer                           |
|      |            $04     4   "I"  Long integer                            |
|      |            $05     8   "$"  currency                                |
|      |            $06     8   "N"  Number                                  |
|      |            $09     1   "L"  Logical                                 |
|      |            $0C     v   "M"  Memo BLOb                               |
|      |            $0D     v   "B"  Binary Large Object                     |
|      |            $0E     v   "F"  Formatted Memo BLOb                     |
|      |            $0F     v   "O"  OLE                                     |
|      |            $10     v   "G"  Graphic BLOb                            |
|      |            $14     4   "T"  Time                                    |
|      |            $15     8   "@"  Timestamp                               |
|      |            $16     4   "+"  Autoincrement                           |
|      |            $17    17*  "#"  BCD                                     |
|      |            $18     v   "Y"  Bytes                                   |
|      |                                                                     |
|      |        The fSize given for BCD fields is not used for field size.   |
|      |        Instead, fSize denotes the number of digits following the    |
|      |        decimal point.  BCD fields are always 17 bytes long.         |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
               end;

    TPxHeader3          =  PACKED RECORD { valid for all table levels }
{==================================================================================================
 offset   type        usage
===================================================================================================}
        recordSize              :  word;
{==============================================================================
| 0000 | int16     recordSize                                                |
|      |                                                                     |
|      |        This is the size of a user record in this table.             |
|      |                                                                     |
|      |        For primary index files, each "record" is actually the       |
|      |        field or fields in the index, plus three int16s which        |
|      |        are not referenced in the header.                            |
|      |                                                                     |
|      |        Secondary index files also have additional fields, but       |
|      |        these are listed in the header.                              |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        headerSize              :  word;
{+------+---------------------------------------------------------------------+
| 0002 | int16     headerSize (always $0800)                                 |
|      |                                                                     |
|      |        You can change headerSize, and move the data blocks          |
|      |        accordingly, to create larger or smaller table headers.      |
|      |        Borland's TUTILITY program would flag an error, but          |
|      |        Paradox, the Borland Database Engine and the Paradox         |
|      |        Engine will all still work with these tables.                |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        fileType                :  byte;
{+------+---------------------------------------------------------------------+
| 0004 | byte        fileType                                                |
|      |                                                                     |
|      |           0 = this is an indexed .DB data file                      |
|      |           1 = this is a primary index .PX file                      |
|      |           2 = this is a non-indexed .DB data file                   |
|      |           3 = this is a non-incrementing secondary index .Xnn file  |
|      |           4 = this is a secondary index .Ynn file (inc or non-inc)  |
|      |           5 = this is an incrementing secondary index .Xnn file     |
|      |           6 = this is a non-incrementing secondary index .XGn file  |
|      |           7 = this is a secondary index .YGn file (inc or non inc)  |
|      |           8 = this is an incrementing secondary index .XGn file     |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        maxTableSize            :  byte;
{+------+---------------------------------------------------------------------+
| 0005 | byte        maxTableSize                                            |
|      |                                                                     |
|      |        This is the "maximum table size" determined when this        |
|      |        table was created.  It really indicates the size of each     |
|      |        block of records in the data section of the table.           |
|      |                                                                     |
|      |           1 =   64M    (block size = $0400 bytes)                   |
|      |           2 =  128M    (block size = $0800 bytes)                   |
|      |           3 =  192M    (block size = $0C00 bytes)                   |
|      |           4 =  256M    (block size = $1000 bytes)                   |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        numRecords              :  int32;
{+------+---------------------------------------------------------------------+
| 0006 | int32     numRecords                                                |
|      |                                                                     |
|      |        This is the number of records in this file.                  |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        nextBlock               :  word;
{+------+---------------------------------------------------------------------+
| 000A | word        nextBlock                                               |
|      |                                                                     |
|      |        I'm not certain what this really is, but it seems to be      |
|      |        the same as fileBlocks unless there is an empty block in     |
|      |        the table.                                                   |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        fileBlocks              :  word;
{+------+---------------------------------------------------------------------+
| 000C | word        fileBlocks                                              |
|      |                                                                     |
|      |        This is the number of data blocks in the file.               |
|      |        (Each "block" is a cluster of records.)                      |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        firstBlock              :  word;
{+------+---------------------------------------------------------------------+
| 000E | word        firstBlock                                              |
|      |                                                                     |
|      |        Always 1 unless the table is empty.                          |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        lastBlock               :  word;
{+------+---------------------------------------------------------------------+
| 0010 | word        lastBlock                                               |
|      |                                                                     |
|      |        This works out to the number of blocks that the table        |
|      |        would contain if every block was packed.                     |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknown12x13            :  word;
{+------+---------------------------------------------------------------------+
| 0012 | word        unknown                                                 |
+------+---------------------------------------------------------------------+}
        modifiedFlags1          :  byte;
{+------+---------------------------------------------------------------------+
| 0014 | byte        modifiedFlags1                                          |
|      |                                                                     |
|      |        A rebuild is required if this is not zero.                   |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        indexFieldNumber	:  byte;
{+------+---------------------------------------------------------------------+
| 0015 | byte        indexFieldNumber                                        |
|      |                                                                     |
|      |        In the .Xnn file of a secondary index, this is the number    |
|      |        of the field it is referencing.                              |
|      |                                                                     |
|      |        This will be zero in the other files.                        |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        primaryIndexWorkspace   :    ptr32;
{+------+---------------------------------------------------------------------+
| 0016 |   pointer     primaryIndexWorkspace                                 |
|      |                                                                     |
|      |        pointer to the primary index file header (in RAM).           |
|      |        This will be a NIL if there is no primary index.             |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknownPtr1A            :    ptr32;
{+------+---------------------------------------------------------------------+
| 001A | pointer     unknown  (suspected pointer)                            |
|      |                                                                     |
|      |        This field is usually a NIL pointer.  I've only seen it      |
|      |        used in 5.0 tables with BCD field types.  It is probably     |
|      |        just a workspace pointer.                                    |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknown1Ex20            :  array[$001E..$0020] of byte;
{+------+---------------------------------------------------------------------+
| 001E...0020        unknown                                                 |
|      |                                                                     |
|      |        I have only seen these three bytes used in .PX files.        |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        numFields               :  int16;
{+------+---------------------------------------------------------------------+
| 0021 | int16     numFields                                                 |
|      |                                                                     |
|      |        This is the number of fields in the table.  If this is an    |
|      |        index file, then it would only be the number of fields in    |
|      |        this index.                                                  |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        primaryKeyFields        :  int16;
{+------+---------------------------------------------------------------------+
| 0023 | int16     primaryKeyFields                                          |
|      |                                                                     |
|      |        This is the number of fields in the file's primary key.      |
|      |        It will be a zero for .PX and .Ynn files; and 2 for .Xnn     |
|      |        secondary index files.                                       |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        encryption1             :  int32;
{+------+---------------------------------------------------------------------+
| 0025 | int32     encryption1                                               |
|      |                                                                     |
|      |        This was where the encryption information was stored for     |
|      |        versions 3.0 and 3.5.  (It was a zero if not encrypted.)     |
|      |                                                                     |
|      |        Subsequent versions store the value $FF00FF00 here, and      |
|      |        move the encryption code to offset $005C.  Even so, these    |
|      |        newer versions still maintain this information at both       |
|      |        locations while working in RAM.                              |
|      |                                                                     |
|      |        Primary and .Ynn secondary index files always use this       |
|      |        field to store the encryption code, but it is often a        |
|      |        zero because the Paradox Engine and the Borland Database     |
|      |        Engine do not always encrypt index files.  You can encrypt   |
|      |        unencrypted index files by following these steps:            |
|      |                                                                     |
|      |           Begin with an empty encrypted table;                      |
|      |           Truncate the index files to headerSize;                   |
|      |           Zeroize nextBlock, fileBlocks, firstBlock and lastBlock;  |
|      |           Copy four bytes from the .DB data file's encryption1      |
|      |               field (for version 3), or the encryption2 field       |
|      |               (for versions 4 and above) into the encryption1       |
|      |               or encryption2 field of the index file;               |
|      |           Test thoroughly.                                          |
|      |                                                                     |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        sortOrder               :  byte;
{+------+---------------------------------------------------------------------+
| 0029 | byte        sortOrder                                               |
|      |                                                                     |
|      |           $00:  ASCII                                               |
|      |           $B7:  International                                       |
|      |           $82:  Norwegian/Danish                                    |
|      |           $E6:  Norwegian/Danish (4.0)                              |
|      |           $F0:  Swedish/Finnish                                     |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        modifiedFlags2          :  byte;
{+------+---------------------------------------------------------------------+
| 002A | byte        modifiedFlags2                                          |
|      |                                                                     |
|      |        A rebuild is required if this is not zero.                   |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknown2Bx2C            :  array[$002B..$002C] of byte;
{+------+---------------------------------------------------------------------+
| 002B...002C        unknown    (always 0)                                   |
+------+---------------------------------------------------------------------+}
        changeCount1            :  byte;
{+------+---------------------------------------------------------------------+
| 002D | byte        changeCount1                                            |
|      |                                                                     |
|      |        This is incremented whenever the file header is updated.     |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        ValSync {originally changeCount2}  :  byte;
{+------+---------------------------------------------------------------------+
| 002E | byte   ValSync.                                                     |
|      |                 Val file synchronization byte. Incremented each     |
|      |        time the VAL file is restructured. This value must match     |
|      |        the first byte of the VAL file, otherwise the engine         |
|      |        reports VAL file out of date error. (Comment by RK)          |
+------+---------------------------------------------------------------------+}
        unknown2F               :  byte;
{+------+---------------------------------------------------------------------+
| 002F | byte        unknown                                                 |
+------+---------------------------------------------------------------------+}
        tableNamePtrPtr         :  pointer;
{+------+---------------------------------------------------------------------+
| 0030 | ^pchar      tableNamePtrPtr                                         |
|      |                                                                     |
|      |        This is a pointer to tableNamePtr, which is a pointer to     |
|      |        tableName.  Paradox uses this field to gain faster access    |
|      |        to tableName because that part of the header is accessed     |
|      |        sequentially.                                                |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        fldInfoPtr              :    ptr32;
{+------+---------------------------------------------------------------------+
| 0034 | pointer     fldInfoPtr                                              |
|      |                                                                     |
|      |        pointer to the list of field identifiers.  This is listed    |
|      |        in the accompanying Pascal record definition as a            |
|      |        PFldInfoRec type.                                            |
|      |                                                                     |
|      |        You can use this pointer value to locate the table header    |
|      |        in memory during run time.  Just subtract $0078 from this    |
|      |        value (for 4.0+ tables), or $0058 (for .PX and .Ynn index    |
|      |        files and version 3.0 tables).                               |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        writeProtected          :  byte;
{+------+---------------------------------------------------------------------+
| 0038 | byte        writeProtected                                          |
|      |                                                                     |
|      |           0        write protection OFF                             |
|      |           1        write protection ON                              |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        fileVersionID           :  byte;
{+------+---------------------------------------------------------------------+
| 0039 | byte        fileVersionID                                           |
|      |                                                                     |
|      |           $03      version 3.0                                      |
|      |           $04      version 3.5                                      |
|      |           $05..09  version 4.x   (usually = $09)                    |
|      |           $0A,$0B  version 5.x                                      |
|      |           $0C      version 7.x                                      |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        maxBlocks               :  word;
{+------+---------------------------------------------------------------------+
| 003A | word        maxBlocks                                               |
|      |                                                                     |
|      |        I don't know what this is for.  It is usually the same       |
|      |        as fileBlocks (at offset 000C).                              |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknown3C               :  byte;
{+------+---------------------------------------------------------------------+
| 003C | byte        unknown                                                 |
+------+---------------------------------------------------------------------+}
        auxPasswords            :  byte;
{+------+---------------------------------------------------------------------+
| 003D | byte        auxPasswords                                            |
|      |                                                                     |
|      |        Number of auxiliary passwords assigned to the table.         |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknown3Ex3F            :  array[$003E..$003F] of byte;
{+------+---------------------------------------------------------------------+
| 003E...003F        unknown                                                 |
+------+---------------------------------------------------------------------+}
        cryptInfoStartPtr       :    ptr32;
{+------+---------------------------------------------------------------------+
| 0040 | pointer     cryptInfoStartPtr                                       |
|      |                                                                     |
|      |        Points to cryptInfo field.  It is always NIL when not        |
|      |        encrypted.  It is sometimes NIL even when encrypted.         |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        cryptInfoEndPtr         :    ptr32;
{+------+---------------------------------------------------------------------+
| 0044 | pointer     cryptInfoEndPtr                                         |
|      |                                                                     |
|      |        Points to end of cryptInfo.  This is NIL if not encrypted.   |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknown48               :  byte;
{+------+---------------------------------------------------------------------+
| 0048 | byte        unknown                                                 |
+------+---------------------------------------------------------------------+}
        autoIncVal              :  int32;
{+------+---------------------------------------------------------------------+
| 0049 | int32     autoInc                                                   |
|      |                                                                     |
|      |        This longint    stores the value used for the next auto      |
|      |        incrementing field in tables with a "+" autoincrementing     |
|      |        field type.                                                  |
|      |                                                                     |
|      |        Formerly used as a modification count.                       |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknown4Dx4E            :  array[$004D..$004E] of byte;
{+------+---------------------------------------------------------------------+
| 004D...004E        unknown                                                 |
+------+---------------------------------------------------------------------+}
        indexUpdateRequired     :  byte;
{+------+---------------------------------------------------------------------+
| 004F | byte        indexUpdateRequired                                     |
+------+---------------------------------------------------------------------+}
(*
        unknown50x54            :  array[$0050..$0054] of byte;
{+------+---------------------------------------------------------------------+
| 0050...0054        unknown                                                 |
+------+---------------------------------------------------------------------+}
    bytes at offset $50..$54 (originally UNKNOWN50x54) do have the following
    interpretation (RK):
 *)
        unknown50               :  byte;
{+------+---------------------------------------------------------------------+
| 0050                unknown                                                 |
+------+---------------------------------------------------------------------+}
        RelevantSpaceSize       : Word;
{+-----+----------------------------------------------------------------------+
| 0051 | word        RelevantSpaceSize                                        |
|      |             This word always stores the number of bytes              |
|      |             in the header that is actually filled with relevant      |
|      |             header information. Remaining bytes (up to HeaderSize)   |
|      |             are usualy filled with zeros (Comment by RK)             |
-------+---------------------------------------------------------------------+}
        unknown53x54            :  array[$0053..$0054] of byte;
{+------+---------------------------------------------------------------------+
| 0053...0054        unknown                                                  |
+------+---------------------------------------------------------------------+}
        refIntegrity            :  byte;
{+------+---------------------------------------------------------------------+
| 0055 | byte        refIntegrity                                            |
|      |                                                                     |
|      |        A value here (=2?) denotes that this table uses              |
|      |        referential integrity checks.                                |
|      |                                                                     |
|      |                                                                     |
|      |             inxDirection (sec'y index file only)                    |
|      |                                                                     |
|      |        Secondary .Xnn index files of v7.0 tables use this           |
|      |        field to indicate sort order direction:                      |
|      |                                                                     |
|      |           $01      ascending sort                                   |
|      |           $11      descending sort                                  |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        unknown56x57            :  array[$0056..$0057] of byte;
{+------+---------------------------------------------------------------------+
| 0056...0057        unknown                                                 |
==============================================================================}
    END;


    TPxHeader4          =  PACKED RECORD { valid for table levels above 4 inclusive }
{==============================================================================
!!!!!!! following part valid only for ver 4.0 or higher     !!!!!!!!!!!!!!!!!!
===============================================================================
 This part of the description applies only to .DB data files and .Xnn index
 files for Paradox versions 4.0 and later :
===============================================================================
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
==============================================================================}
        fileVerID2      :  int16;
{==============================================================================
| 0058 | int16     unknown  (file version ID?)                               |
|      |                                                                     |
|      |           $0105..$0109  version 4.x   (usually = $0109)             |
|      |           $010A, $010B  version 5.x   (usually = $010B)             |
|      |           $010C         version 7.0                                 |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        fileVerID3      :  int16;
{+------+---------------------------------------------------------------------+
| 005A | int16     unknown  (file version ID?)                               |
|      |                                                                     |
|      |        same values as at 0058                                       |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        encryption2     :  int32;
{+------+---------------------------------------------------------------------+
| 005C | int32     encryption2                                               |
|      |                                                                     |
|      |        This will be zero if not encrypted.                          |
|      |        See encryption1 at offset 0025.                              |
|      |        Contains password checksum (comment by RK)                   |
+------+---------------------------------------------------------------------+}
        fileUpdateTime  :  int32;  { 4.0 only }
{+------+---------------------------------------------------------------------+
| 0060 | int32     fileUpdateTime (4.x only)                                 |
|      |                                                                     |
|      |        Format similar to a packed date and time.                    |
|      |        I don't know what this does in 5.0 tables.                   |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        NextInvariantFldNo :  word;  {originally hiFieldID}
{+------+---------------------------------------------------------------------+
| 0064 | word   NextInvariantFldNo                                           |
|      |                                                                     |
|      |        This number is the next Invariant field number to be used    |
|      |        (Comment by RK)                                              |
+------+---------------------------------------------------------------------+}
        InvFldNosOffset {originally hiFieldIDinfo}  : word;
{+------+---------------------------------------------------------------------+
| 0066 | int16     InvFldNosOffset                                           |
|      |                                                                     |
|      |         This word stores the offset of the invariant field          |
|      |         numbers array in the header minus the size of the PxHeader3 |
|      |         part: SizeOf(TPxHeader3)+InvFldNosOffset= Offset of         |
|      |         the first invariant field number         (Comment by RK)    |
-------+---------------------------------------------------------------------+}
        sometimesNumFields:int16;
{+------+---------------------------------------------------------------------+
| 0068 | int16     sometimesNumFields?                                       |
|      |                                                                     |
|      |        This is sometimes the number of fields in the table, but     |
|      |        is often just a zero.  I don't know why.                     |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
        dosCodePage     :  word;
{+------+---------------------------------------------------------------------+
| 006A | int16     dosGlobalCodePage                                         |
|      |                                                                     |
|      |        This was the Global Code Page when this table was created.   |
|      |                                                                     |
|      |           $01B5   United States                                     |
|      |           $02E1   Greek 1                                           |
|      |           $0352   Multilingual (Latin I)                            |
|      |           $0354   Eastern European (Latin II)                       |
|      |           $0359   Turkish                                           |
|      |           $035C   Portuguese                                        |
|      |           $035D   Icelandic                                         |
|      |           $035F   Canadian French                                   |
|      |           $0361   Nordic                                            |
|      |           $0365   Greek 2                                           |
|      |                                                                     |
|      |        Refer to an MS-DOS technical reference about interrupt $21,  |
|      |        function $66, for more information about this.               |
|      |                                                                     |
+------+---------------------------------------------------------------------+}
(*
        unknown6Cx6F    :  array[$006C..$006F] of byte;
{+------+---------------------------------------------------------------------+
| 006C...006F        unknown                                                 |
+------+---------------------------------------------------------------------+}
    bytes at offset 6C..6F (originally unknown6Cx6F) do have the following
    interpretation (RK):
*)
        unknown6Cx6D    :  array[$006C..$006D] of byte;
{+------+---------------------------------------------------------------------+
| 006C...006D        unknown                                                 |
+------+---------------------------------------------------------------------+}
        SortOrderOffset : Word;
{+-----+----------------------------------------------------------------------+
| 006E | word    SortOrderOffset                                             |
|      |                                                                     |
|      |         This word seems to store the offset of the language         |
|      |         driver name in the header minus the size of the PxHeader3   |
|      |         part: SizeOf(TPxHeader3)+SortOrderOffset=offset of          |
|      |         the first language driver name character (Comment by RK)    |
-------+---------------------------------------------------------------------+}
        changeCount4    :  int16;
{+------+---------------------------------------------------------------------+
| 0070 | int16     changeCount4                                               |
+------+---------------------------------------------------------------------+}
        unknown72x77    :  array[$0072..$0077] of byte;
{+------+---------------------------------------------------------------------+
| 0072...0077        unknown                                                  |
==============================================================================}
    END;

    pCharArray = array [1..maxint div sizeOf(pChar)] of pChar;

{ parts of the header }
    T_PxHeaderpart  = (pxh_3,              {header, all levels}
                       pxh_4,              {additional header part, only in levels >= 4}
                       pxh_fieldinfo,      {field type/size info}
                       pxh_createdname,    {created table name}
                       pxh_fieldnames,     {field names}
                       pxh_cryptInfo,      {auxiliary password info}
                       pxh_InvFieldNumbers,{invariant field numbers}
                       pxh_sortorder);     {language driver short name}

    T_PxHeaderparts = set of T_PxHeaderpart;

    T_PxHeader = Class
          PRIVATE
              Header3          : TPxHeader3;
              pHeader4         : ^TPxHeader4;
              pFieldInfo       : ^TPxFieldInfo;
{follows:}
{+======+=====================================================================+
|      | pTableName       : pChar;                                           |
|      |                                                                     |
|      |       Pointer to tableName when header is in RAM.                   |
|      | Comment by RK   - NOT stored in the class                           |
+======+=====================================================================+
|      |  array[1..(numFields)] of pchar                                     |
|      |                                                                     |
|      |      * These pointers are not present in the index files. *         |
|      |                                                                     |
|      |        This is an array of pointers that reference the field names  |
|      |        when Paradox (or one of the engines) is running.  The size   |
|      |        of this array depends upon the number of fields.             |
|      | Comment by RK   - NOT stored in class                               |
+======+=====================================================================+}
              pCreatedTableName : pChar;
{==============================================================================
|      |        This was the name this file was assigned when created.       |
|      |        It will use 79 bytes, padded with zeroes.                    |
|      |                                                                     |
|      |        You will often find the filename "RESTTEMP.DB" here.         |
|      |                                                                     |
|      |  Comment by RK    - FIXED LENGTH ASCIIZSTRING                       |
|      |        levels prior to 7 use 79 chars including terminating #0      |
|      |        level 7 uses 261 chars including terminating #0              |
==============================================================================}
              pPHYSFieldNames  : ^pCharArray;
{------+---------------------------------------------------------------------+
|      |        These are the ASCIIZ field name(s), arranged sequentially.   |
+------+---------------------------------------------------------------------+}
              pCryptInfo       : pointer;
{+------+---------------------------------------------------------------------+
|      |                                                                     |
|      |        Encrypted tables would have additional data here.            |
|      |                                                                     |
|      | Comment by RK -  Tables with auxiliary passwords would have 256     |
|      |        + Number of fields bytes here                                |
+------+---------------------------------------------------------------------+}
               pInvariantFieldNumbers: ^twordarray;
{+------+---------------------------------------------------------------------+
| ---- |  array[1..(numFields)] of word     InvariantFieldNumbers            |
|      |                                                                     |
|      |        Invariant field numbers (Comment by RK)                      |
+------+---------------------------------------------------------------------+}
               pSortOrderId    : pChar;
{==============================================================================
| ---- | char[]      sortOrderID                                             |
|      |                                                                     |
|      |        An ASCIIZ string representing the sort order for this        |
|      |        table ("ascii", "intl", etc.).                               |
|      |                                                                     |
==============================================================================}
               Function    GetTableLevel: word; {level*10}
               Function    GetFieldCount: Int16;
               { Returns size of space reserved for create-time table name
                 in the header. See "pCreatedTableName" field :}
               Function    GetCreatedTableNameSpace: word;
               { Returns size of the space used for storing auxiliary
                 password information }
               Function    GetCryptInfoSize : int16;
               Function    GetHasAuxPasswords: boolean;
               { Returns master password checksum }
               Function    GetMpswCheckSum: longint;
           PUBLIC
               Constructor Create(const DBFile: String);
               Destructor  Destroy; override;
               { writes parts of the header to file: }
               Procedure   WriteHeaderParts(const DBFile: string; parts: T_PxHeaderparts);
               Property    Tablelevel: word {level*10} read GetTableLevel;
               Property    HasAuxPasswords: Boolean read GetHasAuxPasswords;
               Property    MasterPasswordChecksum: int32 read GetMpswCheckSum;
               Property    ValSyncByte: byte read Header3.ValSync;
               Property    AutoIncLong: int32 read Header3.autoIncVal write Header3.autoIncVal;
               Property    CountRecords: int32 read Header3.NumRecords;
    END;

    EInvalidHeader = class(Exception)
    end;

CONST
    unsupportedlevel = 65535; { unsupported table level tag }

implementation
{all comments below by RK}

Constructor T_PxHeader.Create(const DBFile: String);
{ Exception within this constructor indicates "Header inconsistent"
  corruption. The user will be given the option to borrow the
  header from the backup table }
var i,j: integer;
    thebuff: array [1..256] of byte;
begin
  Try
    inherited Create;
    with tFileStream.Create(DBFile,fmOpenRead or fmShareDenyNone) do TRY
{Header3}
      Read(Header3,SizeOf(TPxHeader3));
{Header4}
      if TableLevel < unsupportedlevel then begin
        if TableLevel >= 40 then begin
          GetMem(pHeader4,SizeOf(TPxHeader4));
          Read(pHeader4^,SizeOf(TPxHeader4));
        {Check pHeader4^.NextInvariantFldNo consistency }
          if pHeader4^.NextInvariantFldNo <= GetFieldCount
              then sysutils.abort; {leads to EInvalidHeader exception }          
        end;
  {FieldInfo}
        GetMem(pFieldInfo,SizeOf(TPxFieldInfo)*GetFieldCount);
        Read(pFieldInfo^,SizeOf(TPxFieldInfo)*GetFieldCount);
  {pTablename - not stored in class }
        Seek(SizeOf(pChar),soFromCurrent);
  {RAM field names - not stored in class }
        Seek(GetFieldCount*SizeOf(pChar),soFromCurrent);
  {CreatedTablename}
        GetMem(pCreatedTableName,GetCreatedTableNameSpace);
        Read(pCreatedTableName^,GetCreatedTableNameSpace);
  {pPHYSFieldNames}
        pPHYSFieldNames := AllocMem(SizeOf(pChar)*GetFieldCount);
        For i := 1 to GetFieldCount do
          For j := 1 to SizeOf(TheBuff) do begin
            Read(thebuff[j],1);
            if thebuff[j] = 0 then begin
              GetMem(pPHYSFieldNames^[i],j);
              Move(thebuff,pPHYSFieldNames^[i]^,j);
              break; {break inner loop }
            end;
          end;
  {Crypt info}
        if GetCryptInfoSize > 0 then begin
          GetMem(pCryptInfo,GetCryptInfoSize);
          Read(pCryptInfo^,GetCryptInfoSize);
        end;
        if TableLevel >= 40 then begin
    {Check pHeader4^.InvFldNosOffset consistency: }
          if Position <> (pHeader4^.InvFldNosOffset + SizeOf(TPxHeader3))
              then sysutils.abort; {leads to EInvalidHeader exception }
    {Invariant field numbers }
          GetMem(pInvariantFieldNumbers,GetFieldCount*SizeOf(Word));
          Read(pInvariantFieldNumbers^,GetFieldCount*SizeOf(Word));
    {Check invariant field numbers consistency: }
          For j := 0 to GetFieldCount-1 do begin
            if pInvariantFieldNumbers^[j] >= pHeader4^.NextInvariantFldNo
              then sysutils.abort; {leads to EInvalidHeader exception }
          end;
    {Check the pHeader4^.SortOrderOffset consistency: }
          if Position <> (pHeader4^.SortOrderOffset + SizeOf(TPxHeader3))
              then sysutils.abort; {leads to EInvalidHeader exception }
    {SortOrderId}
          For j := 1 to SizeOf(TheBuff) do begin
            Read(thebuff[j],1);
            if thebuff[j] = 0 then begin
              GetMem(pSortOrderId,j);
              Move(thebuff,pSortOrderId^,j);
              break;
            end;
          end;
        end;
      end;
      if Position <> Header3.RelevantSpaceSize {header is corrupt}
        then sysutils.abort; { this leads to EInvalidHeader exception }
    Finally
      Free;
    end;
  Except
    raise EInvalidHeader.Create('invalid header');
    {message will never appear, no need to translate}
  End;
end;

Destructor  T_PxHeader.Destroy;
Var i: integer;
begin
  if pSortOrderId <> nil then FreeMem(pSortOrderId,StrLen(pSortOrderId)+1);
  if pInvariantFieldNumbers <> nil then FreeMem(pInvariantFieldNumbers,GetFieldCount*SizeOf(Word));
  if pCryptInfo <> nil then FreeMem(pCryptInfo, GetCryptInfoSize);
  if pCreatedTableName <> nil then FreeMem(pCreatedTableName,GetCreatedTableNameSpace);
  if pPHYSFieldNames <> nil then begin
    For i := 1 to GetFieldCount do
      FreeMem(pPHYSFieldNames^[i],StrLen(pPHYSFieldNames^[i])+1);
    FreeMem(pPHYSFieldNames,SizeOf(pChar)*GetFieldCount);
  end;
  if pFieldInfo <> nil then FreeMem(pFieldInfo,SizeOf(TPxFieldInfo)*GetFieldCount);
  if pHeader4 <> nil then FreeMem(pHeader4,SizeOf(TPxHeader4));
  inherited Destroy;
end;

procedure   T_PxHeader.WriteHeaderParts(const DBFile: string; parts: T_PxHeaderparts);
var i: integer;
begin
  with tFileStream.Create(DBFile,fmOpenWrite) do TRY
    if pxh_3 in parts then
      Write(Header3,SizeOf(Header3))
    else if pxh_cryptInfo in Parts then begin
     { Write Header3.auxPasswords - offset $3D }
      Seek($3D,soFromCurrent);
      Write(Header3.auxPasswords,1);
     { Write Header3.RelevantSpaceSize - offset $51 }
      Seek($51-$3D-$01,soFromCurrent);
      Write(Header3.RelevantSpaceSize,2);
      Seek(SizeOf(Header3)-$51-$02,soFromCurrent);
    end else
      Seek(SizeOf(Header3),soFromCurrent);
{}  Exclude(parts,pxh_3);
{}  if parts = [] then exit;
    if pxh_4 in parts then
      Write(pHeader4^,SizeOf(pHeader4^)) {raises exception if pHeader4 = nil}
    else if (Tablelevel >= 40) and ([pxh_cryptInfo,pxh_InvFieldNumbers]*Parts <> []) then begin
      Seek($64-SizeOf(Header3),soFromCurrent);
      if pxh_InvFieldNumbers in Parts then
       { Write pHeader4^.NextInvariantFldNo - offset $64 }
        Write(pHeader4^.NextInvariantFldNo,2)
      else
        Seek(2,soFromCurrent);
      if pxh_cryptInfo in Parts then begin
     { Write pHeader4^.InvFldNosOffset - offset $66 }
        Write(pHeader4^.InvFldNosOffset,2);
     { Write pHeader4^.SortOrderOffset - offset $6E }
        Seek($6E-$66-$02,soFromCurrent);
        Write(pHeader4^.SortOrderOffset,2);
        Seek(SizeOf(pHeader4^)+SizeOf(Header3)-$6E-$02,soFromCurrent);
      end else
        Seek(SizeOf(pHeader4^)+SizeOf(Header3)-$64-$02,soFromCurrent);
    end else
      Seek(SizeOf(pHeader4^),soFromCurrent);
{}  Exclude(parts,pxh_4);
{}  if parts = [] then exit;
    if pxh_fieldinfo in parts then
      Write(pFieldInfo^,SizeOf(TPxFieldInfo)*GetFieldCount)
    else
      Seek(SizeOf(TPxFieldInfo)*GetFieldCount,soFromCurrent);
{}  Exclude(parts,pxh_fieldinfo);
{}  if parts = [] then exit;
  {pTablename - not stored in class }
    Seek(SizeOf(pChar),soFromCurrent);
  {RAM field names - not stored in class }
    Seek(GetFieldCount*SizeOf(pChar),soFromCurrent);
    if pxh_createdname in parts then
      Write(pCreatedTableName^,GetCreatedTableNameSpace)
    else
      Seek(GetCreatedTableNameSpace,soFromCurrent);
{}  Exclude(parts,pxh_createdname);
{}  if parts = [] then exit;
    if pxh_fieldnames in parts then
      For i := 1 to GetFieldCount do
        Write(pPHYSFieldNames^[i]^,StrLen(pPHYSFieldNames^[i])+1)
    else
      For i := 1 to GetFieldCount do
        Seek(StrLen(pPHYSFieldNames^[i])+1,soFromCurrent);
{}  Exclude(parts,pxh_fieldnames);
{}  if parts = [] then exit;
    if pxh_cryptInfo in parts then
      Write(pCryptInfo^, GetCryptInfoSize)
    else
      Seek(GetCryptInfoSize,soFromCurrent);
{}  Exclude(parts,pxh_cryptInfo);
{}  if parts = [] then exit;
    if pxh_InvFieldNumbers in Parts then
      Write(pInvariantFieldNumbers^,GetFieldCount*SizeOf(Word))
    else
      Seek(GetFieldCount*SizeOf(Word),soFromCurrent);
{}  Exclude(parts,pxh_InvFieldNumbers);
{}  if parts = [] then exit;
    if pxh_sortorder in parts then begin
      Write(pSortOrderId^,StrLen(pSortOrderId)+1)
    end else begin
      Seek(StrLen(pSortOrderId)+1,soFromCurrent);
    end;
  Finally
    Free;
  End;
end;

Function    T_PxHeader.GetTableLevel: word; {level*10}
begin
   if Header3.FileVersionId <= $03  then Result := 30
   else
   if Header3.FileVersionId <= $04  then Result := 35
   else
   if Header3.FileVersionId <= $09  then Result := 40
   else
   if Header3.FileVersionId <= $0B  then Result := 50
   else
   if Header3.FileVersionId <= $0C  then Result := 70
   else { don't have info about }
     Result := unsupportedlevel; { unsupported }
end;

Function    T_PxHeader.GetFieldCount: int16;
begin
  Result := Header3.numFields;
  if (Result<=0) or (Result >=256) then sysutils.abort; {leads to Einvalid header}
end;

Function    T_PxHeader.GetCryptInfoSize : int16;
begin
  if HasAuxPasswords then
    Result := 256+GetFieldCount
  else
    Result := 0;
end;

Function    T_PxHeader.GetCreatedTableNameSpace: word;
begin
  if TableLevel >= 70 {level 7} then Result := 261
  else Result := 79;
end;

Function    T_PxHeader.GetHasAuxPasswords: boolean;
begin
  Result := Header3.auxPasswords > 0
end;

Function T_PxHeader.GetMpswCheckSum: int32;
begin
  if TableLevel < 40 then Result := Header3.encryption1
  else Result := pHeader4^.Encryption2;
end;


end.
