{=====================================================================
 Pdxrbld - Paradox tables repair utility  Copyright 1998 by
 Roman Krejci-RKS, Pujmanove 1552, 140 00 Prague 4, Czech Republic
 http://www.rksolution.cz, EMail INFO@rksolution.cz
 =====================================================================}

unit Rbld1;
{$I PDXRBLD.INC}

interface

uses
{$IFDEF WIN32}
  Windows,
  BDE, TUtil32,
{$ELSE}
  Wintypes, Winprocs, dbiErrs, dbiTypes, dbiProcs,
  TUtility,
{$ENDIF}
  SysUtils, Classes, DB, DBTables, Filectrl, Controls, Dialogs, RKSIni,
{$IFNDEF WIN32}
  VarTbl, { "Variant" type and tTable with Locate method }
{$ENDIF}
  PxHeader,Translat;

const
 { DATA parameter values in GENPROGRESS callbacks: }
  clbdataTUTILITYVerify      = 0;  { callback from within  TUTILITY - verify }
  clbdataTUTILITYRebuild     = 1;  { callback from within  TUTILITY - rebuild }
  clbdataPDXRBLD             = 2;  { callback from within this unit }
  clbdataPack                = 3;  { callback from within  BDE (pack process) }
  clbdataRestructure         = 4;  { callback from within  BDE (other restructure process) }

 { error codes }
  ErrCodeUnreparableError    = 200;
  ErrCodeVALOutOfDate        = 201;
  ErrCodeCorruptedAutoinc    = 202;
  ErrCodeOrphansFound        = 203;
  ErrCodeUnsupportedVersion  = 204;

 { size of Paradox table header }
  PxHeaderSize               = $800;

 { rebuild scope }
  rsNone      = 0;
  rsCorrupted = 1;
  rsAll       = 2;
  rsAllSkipTUVerify = 3;


Type
{ Callbacks: }
  TGenProgressCallBack = function (ecbType: CBType; Data: LongInt; pcbInfo: Pointer): CBRType;
   {$IFDEF WIN32}
     stdcall;
   {$ENDIF}

  TRebuildProcess = (cpVerify,cpRepair,cpPack,cpOrphansCheck);
  TCallbackState  = (cstStarting,cstReportingErrors,cstReportingResult);

  TSelectDBFolderFunc = function (Const CorruptDB,CorruptionDescr: String): String;
{ Function returning path to backup database from which the table/index header
  should be borrowed. Should return empty string if no database was selected.
  CorruptDB is name of the corrupted table.
  CorruptionDescr is dbi error message
**Change in version 3.00 - function should now return path to backup database
  (folder only). In previous versions, it returned path to DB file !!! }

  PCallbackRecord = ^TCallbackRecord;
  TCallbackRecord = packed record
    percentdone : smallint;
    aProcess    : TRebuildProcess;
    aState      : TCallbackState;
{ VERIFY.DB structure: }
    Drive       : string;
    Directory   : string;
    tablename   : string;
    Extension   : string;
    ErrorCode   : Smallint;
    Errormessage: string;
    date        : tdateTime;
    Time        : tdateTime;
  end;

  tCBRESTRPROGRESSDesc = packed record
  { first two fields are CBPROGRESSDesc record }
    iPercentDone: smallint;
    szMsg       : DBIMSG;
    TotalRecords: longint;
  end;


Procedure CheckAndRepairTables(
               const datBname: string;     {databasename}
               IsAlias: boolean;           {datBname is Alias (as opposed to Path) }
               scopeRepair: integer;       {don't repair, repair if corrupted , repair all,
                                            repair all and skip TUVerifytable step }
               DoPack: Boolean;            {pack tables}
               TblCallback,                {TUTILProgressCallBack - progress on current table. can be NIL. }
               DatabaseCallback: pointer;  {TUTILProgressCallBack - progress on entire database. Can be NIL}
               const tableSelected:String; {Table to check/repair. Empty string=ALL}
               SelectDBFolderFunc:TSelectDBFolderFunc); { Function returning file for header borrow. Can be NIL }

{$IFDEF WIN32}
Procedure AddSessionPassword(const Px3PKey,Pwd: string);
Procedure RemovePx3PKey(const Px3PKey: string);
Procedure RemoveAllPx3PKeys;
{$ELSE}
Procedure AddSessionPassword(const Pwd: string);
{$ENDIF}
{ Adds password to the session. At the same time updates application password list }

implementation

{$IFDEF WIN32}
{$R P.R32}
{$ELSE}
{$R P.R16}
{$ENDIF}

USES LowStuff,PSWD;

{ TUtility error codes (field ErrorCode in Verify.DB)
 -----------------------------------------------------
 1	Unable to open file %s
 2	Invalid Header size %u
 3	Invalid file size %ld: %s
 4	Invalid block size %u: %s
 5	Real header size %u is invalid
 6	Invalid sort byte in header
 7	Sum of field sizes does not equal record size
 8	Not enough memory to complete operation
 9	Header size is not a multiple of 2K
 10	Unsupported field type in field %u
 11	Field name greater than 25 char
 12	Header real size %u is greater than start of data %u
 13	File type %s does not match extension %s
 14	Unknown file type
 15	Keyed db missing px file.
 16	Maintained index missing .y* file.
 17	Invalid record size
 18	Unable to read file %s
 19	Table header reports damage
 20	Total blocks (%u) greater than max blocks (%u)
 21	Header inconsistent
 22	Invariant Field ID %u for Field %u is greater than max field id %u
 23	First block greater than total blocks
 24	Last block greater than total blocks
 25	Index version does not match table version
 26	Index record size does not match table keysize
 27	Block %u is in the chain twice
 28	Number of records %i in block %u  does not match index %li
 29	Reported header size is not large enough to hold header information
 30	Header size (%u) + data size (%ld) does not equal filesize (%ld)\nDifference: %ld bytes
 31	Block pointers in block %u do not match pointers in block %u
 32	First record of Block %u does not match index
 33	Forward block pointer in block %u is out of range
 34	Number of records in index file is greater than the number of blocks in data file
 35	Record %lu is not in sort order
 36	Empty block (%u) in used block chain
 37	Number of blocks is greater than the number of records in index
 38	Last block in chain (%u) does not match header (%u)
 39	Number of records found (%lu) does not match number of header records (%lu)
 40	Number of blocks in used block chain (%u) does not equal used blocks in header (%u)
 41	Empty block %u reports records
 42	Empty block pointer in block %u is out of range
 43	Total Blocks reported by header (%u) does not equal blocks found in chain (%u)
 44	Previous block pointer in block %u is out of range
 45	Secondary Index %s does not match table
 46	Used blocks greater than total blocks\n\tUsed blocks = %u\n\tTotal blocks = %u
 47	First block = 0 and header reports used blocks
 48	Data start (%u) is greater than file size (%lu)
 49	Hanging .PX file
 50	Hanging .Y** file
 51	Table is full.  Unable to allocate block
 52	BTree record mismatch from block (%u) record(%i) to block (%u)
 53	BTree record number mismatch from block (%u) record(%i) to block (%u)
 54	Pointer to last record (%i) in Block %u is not consistent with record size (%u)
 55	Alpha Field %s in record %lu contains low ascii chars
 56	Field %s of record %lu does not match secondary index %s
 57	Field number out of bounds in secondary index %s
 58	Invalid number of fields.
 59	Field size %u is invalid for field type %s in field number %u
 60	Invalid table format
 61	Block %u points to itself
 62	Number of key fields is greater than the number of fields in the table
 63	Write-protect flag
 64	Sort name missing from header
 65	The indices are out of date:\nA rebuild is recommended!
 66	Modified flags are set:\nA rebuild is recommended!
 67	Invalid memo field
 68	Header not consistent with format
 69	Header inconsistent with version
 70	Invalid sequence number
 71	Data in field %s of record %lu does not match .MB file
 72	Blob pointer offset for field %u of record %lu points to a memo block header of wrong or invalid type
 73	Blob pointer index for field %u of record %lu is invalid
 74	Blob pointer length for field %u of record %lu does not match .MB file
 76	Memo file signature is not valid
 75	Blob pointer sequence number for field %u of record %lu does not match .MB file
 77	Unsupported memo version
 78	Unsupported memo free block size
 79	Unsupported memo increment size
 80	Unsupported memo list block size
 81	Unsupported memo small block index size
 82	Unsupported memo big block size
 83	Unsupported memo small block size
 84	Missing .MB file
 85	File size is too large to be a replica table
 86	Replica table reports records
 87	Float field %s in record %lu has denormal value
 88	Float field %s in record %lu has exponent smaller than accepted minimum
 89	Float field %s in record %lu has +/- infinity value
 90	Float field %s in record %lu has quiet NAN value
 91	Float field %s in record %lu has signaling NAN value
 92	Float field %s in record %lu has exponent larger than accepted maximum
 93	Used memo block in free list chain
 94	Free memo size does not match size reported in free list
 95	Language driver unsupported
 96	Invalid formatted memo header
 97	Secondary Index %s is out of date


 Some errors are known to be unreparable: }

const
   { "Header inconsistent" error will be reported in
     case Pdxrbld cannot succesfuly read important
     info in the DB file header. }
    HeaderInconsistent = 21;

    UnreparableErrors: Set of byte = [
        1, {unable to open file}
        8, {not enough memory}
       14, {unknown fileType}
       18, {unable to read file}
       63, {write protect flag}
       77..83, {unsupported version of ...}
       84, {missing MB file }
       95  {language driver unsupported }
       ];

    MissingPXFile = 15; { Keyed DB missing PX file - Removed from UnreparableErrors in ver 2.60 }
    MissingYFile  = 16; { Maintained index missing .Y* file - Removed from UnreparableErrors in ver 2.60 }

    { Some corruptions should be fixed by restructure (pack) only - TUTILITY rebuild would turn
      data unreadable. Table will be unreparable if it contains errors that should be fixed by
      restructure only mixed with other types of errors, with the exception of
      "RuleExceptionCombinations" }
    FixByRestructureErrors: Set of byte = [
      28, {Number of records %i in block %u  does not match index %li
           fix for 4.20 :
           ..George Guorgi 19.9.2000   [28,30] combination IS repairable by restructure only }
      30  {Header size (%u) + data size (%ld) does not equal filesize (%ld)\nDifference: %ld bytes}
     ];

    { Note: Combination [30,53], CAN be successfuly repaired by TUTILITY rebuild.
            Daniel Basinger, 19.9.2000. fix for 4.20 }
    { 53: BTree record number mismatch from block (%u) record(%i) to block (%u) }
    RuleExceptionCombination1 = [30, 53];

   { Some corruptions of *.X* index file require header replacement }
    XIndexReplaceHeaderErrors = [
     2, {Invalid Header size %u}
     4, {Invalid block size %u: %s}
     5, {Real header size %u is invalid}
     6, {Invalid sort byte in header}
     9, {Header size is not a multiple of 2K}
    11, {Field name greater than 25 char}
    12, {Header real size %u is greater than start of data %u}
    19, {Table header reports damage}
    21, {Header inconsistent}
    29, {Reported header size is not large enough to hold header information}
    45, {Secondary Index %s does not match table}
    57, {Field number out of bounds in secondary index %s}
    58, {Invalid number of fields.}
    64, {Sort name missing from header}
    68, {Header not consistent with format}
    69  {Header inconsistent with version}
     ];

    FormattedMemoError = 96; { Invalid formatted memo header }
    { This by itself is not an error, though TUTility reports it as this.
      Rebuild routine should be avoided if FormattedMemoError is reported,
      because rebuild would effectively erase all formatted memo data.
      Pdxrbld uses this approach when formatted memo is present:
         FormatedMemoError is simply ignored and no error is reported.
         If, however, TUTILITY rebuild is required either because the
         table contains additional errors or the user selected "rebuild all"
         rebuild scope, request for Rebuild action will fail with
         "Cannot rebuild formatted memo" error message.

{ Unfortunately, I was unable to obtain tables with complete sample of the above
  listed corruptions. If you ever discover that PdxRbld does not correctly repair
  an error, please extend the  UnreparableErrors set. If that happens, I would be
  most grateful for receiving a note on INFO@rksolution.cz (or KREJCI@mbox.cesnet.cz) }

{ Orphans checking }
  iniOcheck  ='OCHECK.INI'; { INI file in data folder }
  iniMaster  ='MASTER';     { base value name for master table, padded with integer in the INI file (MASTER1 etc) }
  iniMfields ='MFIELDS';    { base value name for master fields, padded with integer in the INI file (MFIELDS1 etc) }
  iniDfields ='DFIELDS';    { base value name for detail fields, padded with integer in the INI file (DFIELDS1 etc) }

{ Orphans detection/removal feature was added to version 2.0.
  Information about linked tables and linked field could be obtained automatically
  if each master detail link is accopagnied by the corresponding referential integrity
  rule. Referential integrity, however, imposes significant performance load on the application,
  so one can often see master detail links with no associated referential integrity.
  In addition, circular master/detail links are possible in Delphi (A->B link in one datamodule,
  B->A link in another datamodule) whereas circular referential integrity is not allowed.
  From this follows that automatic link detection is not possible and the link description
  must be provided externally. Pdxrbld uses INI file for this purpose. If you want to use
  orphans detection/removal feature, place file named OCHECK.INI into data folder. The structure
  of the INI file should be as follows:

  Section names are simply detail table names (without extension!).
  Each section can hold any number of triplets

  MASTERx=Master_Table
  MFIELDSx=Master_Fields
  DFIELDSx=Detail_Fields

  where x is the triplet number (must start from 1 and be consecutive),
  Master_Table is name of the master table (with or without extension),
  Master_Fields is the semicolon separated list of master table's field names
               involved in the link (as they appear in the MasterFields tTable property)
  Detail_Fields is the semicolon separated list of detail table's field names
               involved in the link (as they appear in the IndexFieldNames tTable property)

  !!! Detail table *MUST* have primary index, otherwise an exception is raised !!!

  Example of OCHECK.INI for DBDEMOS alias follows:

---- CUT HERE---------------
[ITEMS]
MASTER1=PARTS
MFIELDS1=PartNo
DFIELDS1=PartNo
MASTER2=ORDERS
MFIELDS2=OrderNo
DFIELDS2=OrderNo

[PARTS]
MASTER1=VENDORS
MFIELDS1=VendorNo
DFIELDS1=VendorNo

[ORDERS]
MASTER1=CUSTOMER
MFIELDS1=CustNo
DFIELDS1=CustNo
---- END CUT ---------------

If OCHECK.INI is not present in the data folder, orphans detection/removal is skipped.

}

{ table names of auxiliary tables in private directory }
  backupdb  = 'BACKUP.DB';
  keyvioldb = 'KEYVIOL.DB';
  verifydb  = 'VERIFY.DB';
  rebuilddb = 'REBUILD.DB';
  problemdb = 'PROBLEM.DB';

{ field names in Verify table (VERIFY.DB in private directory) }
{ Loaded from TUTILITY string resources at runtime. Numbers are resource-ID's }
  TabVerifyDrive         = 512; {'Drive'}
  TabVerifyDirectory     = 513; {'Directory'}
  TabVerifyTable_name    = 514; {'Table name'}
  TabVerifyExtension     = 515; {'Extension'}
  TabVerifyDate          = 519; {'Date'}
  TabVerifyTime          = 520; {'Time'}
  TabVerifyError_code    = 516; {'Error code'}
  TabVerifyError_level   = 517; {'Error level'}
  TabVerifyError_message = 518; {'Error message'}
  
{ various messages for programmer errors that need not to be localised }
  { error in OCHECK.INI file }
  InvalidOCheckINI = 'Invalid file %s for detail %s: '#13'Master=%s'#13'Master fields=%s'#13'Detail fields=%s';
  { primary index required }
  priIxRequired    = 'Orphans detection/removal requires primary index (table %s)';


type
  TBDEUtil = class;

  TRbldClass1 = class
    TabVerify: TTable;
    BDEUtil  : TBDEUtil;
    Constructor Create;
    Destructor Destroy; override;
    Function VerifyTable(Const Table: String; Var Msg: Integer; const pswd: string): DBIResult;
    Function Rebuildtable(Const szTable,pswd: String): DBIResult;
{$IFDEF WIN32} {OnPassword is *not* an event in 16 bit Delphi }
    procedure NoDefaultpassword(Sender: TObject; var Continue: Boolean);
      { Session's Onpassword event - we need to bypass the default tSession password dialog }
{$ENDIF}
  end;

  TBDEUtil = class
    CbInfo: TUVerifyCallback;
    TUProps: CURProps;
    hDb: hDBIDb;
    vhTSes: hTUSes;
    constructor Create;
    destructor Destroy; override;
    procedure RegisterCallBack(Clback: Pointer; clbID: longint);
    procedure UnRegisterCallBack(clbID: longint);
  end;

  TPendingOrphan = class        { holds information that identifies orphaned record }
    TableName    : string;
    PriIXFields  : tStringList; { each item holds name of one primary index field }
    OrphanedValue: variant;     { Variant array. OrphanedValue[i] holds Table[PriIXFields[i]] value of the orphaned record }
    PendingInPass: Integer;     { CurrentPassNumber when instance was created }
    Constructor Create(tbl: tTable; aPriIXFields: tStringList);
    Destructor  Destroy; override;
    Function    CompareRecord(tbl: tTable): boolean;  { tells whether instance match current record in tbl }
    Function    CompareInstance(Inst: TPendingOrphan): boolean; { tells whether instance match Inst }
  end;

  TPendingOrphans = class(tList)                                 { list of TPendingOrphan }
    Destructor Destroy; override;
    Procedure  ClearItems;                                       { Frees all items and clears the list }
    Function   FindRecord(tbl: tTable): TPendingOrphan;          { Finds item matching the current record in tbl }
    Procedure  AddRecord(tbl: tTable; aPriIXFields: tStringList);{ Adds an item, checks for duplicates }
  End;

VAR
{$IFDEF WIN32}
  Px3PKeysList  : tStringList; { List of Px3P keys/passwords added to the session }
                               { strings=Px3P keys, objects=tStringlist holding:
                                                    strings= password
                                                    objects= password checksum under
                                                             particular Px3P key }
{$ELSE}
  PswdList : tStringList; { strings= password, objects= password checksum  }
{$ENDIF}

  PendingOrphans: TPendingOrphans;
{
  Orphans detection and removal is implemented quite simply:
  Detail table is iterated and for each detail record corresponding master from
  master table is searched for. If the routine does not find master record,
  detail record (orphan) is deleted if repair of damaged tables is allowed.
  It may happen, however, that orphaned record cannot be deleted because there
  are detail records in some other table that reference this very orphan as master
  through some other referential integrity relationship. The records in other table
  themselves are *NOT* orphans - because their master exists - and they would
  never be recognized as superflous, and, therefore, would never be deleted.
  Thought it is basically possible to solve this problem by querying referential integrity
  descriptions of all tables in the database (dbiGetErrorContext does not provide
  any information about offending table after DBIERR_DETAILRECORDSEXIST BDE exception
  has been raised!), the actual implementation would be too complicated.
  Here we use different approach: if the orphan cannot be deleted (due to the referential
  integrity constraint, for example) primary key value for orphaned record is stored in
  suitable object (tPendingOrphan instance) and this object is placed into the list
  of pending orphans.
  Record is then detected as orphaned if any of following two condition holds:
     1) master record cannot be found,
     2) master record match any item in the pending orphans list

  Because the number of detected orphans might depend on the order in which the tables
  are checked, pending orphans list is made global so it can hold the information
  between subsequent Checks/Rebuilds of the same database.

  This logic allows to remove all orphans in a finite sequence of Checks/Rebuilds
  of the same database.

  >> If any of the error messages states that some orphans could not be deleted,
     run the repair/check again immediately on the same database <<

  If circular links are present, it may be necessary to repeat Check/Rebuild
  even if the message states that all orphans were succeesfuly removed,
  until no error message appears.
}

  CurrentPassNumber : Integer; { Number of subsequent passes through the same database }
  LastDataFolder    : String;  { Holds path to last checked database.
                                 if database changes, PendingOrphans is cleared and
                                 CurrentPassNumber is set to 1 }
{ TPendingOrphan }

Constructor TPendingOrphan.Create(tbl: tTable; aPriIXFields: tStringList);
var i: integer;
begin
  tablename      := AnsiUppercase(tbl.tablename);
  PriIXFields    := tStringList.Create;
  PriIXFields.Assign(aPriIXFields);
  PendingInPass  := CurrentPassNumber;
{$ifdef win32}
  OrphanedValue  := VarArrayCreate([0, PriIXFields.Count-1], varVariant);
{$else}
  OrphanedValue  := VarArrayCreate([0, PriIXFields.Count-1]);
{$endif}
  for i := 0 to PriIXFields.Count-1 do OrphanedValue[i] := tbl[PriIXFields[i]];
end;

Destructor  TPendingOrphan.Destroy;
begin
{$ifndef win32}
  OrphanedValue.Free;
{$endif}
  PriIXFields.Free;
  inherited Destroy;
end;

Function TPendingOrphan.CompareRecord(tbl: tTable): boolean;
var i: integer;
begin
  Result := false;
  if tablename = AnsiUppercase(tbl.tablename) then begin
    for i := 0 to PriIXFields.Count-1 do
{$IFDEF WIN32}
    if OrphanedValue[i] <> tbl[PriIXFields[i]] then exit;
{$ELSE}
    if not EqualVariants(OrphanedValue[i],tbl[PriIXFields[i]]) then exit;
{$ENDIF}
    Result := true;
  end;
end;

Function TPendingOrphan.CompareInstance(Inst: TPendingOrphan): boolean;
var i: integer;
begin
  Result := false;
  if tablename = Inst.tablename then begin
    for i := 0 to PriIXFields.Count-1 do
{$IFDEF WIN32}
    if OrphanedValue[i] <> Inst.OrphanedValue[i] then exit;
{$ELSE}
    if not EqualVariants(OrphanedValue[i],Inst.OrphanedValue[i]) then exit;
{$ENDIF}
    Result := true;
  end;
end;

{ TPendingOrphans }

Destructor TPendingOrphans.Destroy;
begin
  ClearItems;
  inherited Destroy;
end;

Procedure TPendingOrphans.ClearItems;
var i: integer;
begin
  for i := 0 to count-1 do TPendingOrphan(Items[i]).Free;
  Clear;
end;

Function   TPendingOrphans.FindRecord(tbl: tTable): TPendingOrphan;
var i: integer;
begin
  Result := NIL;
  for i := 0 to count-1 do
    if TPendingOrphan(Items[i]).CompareRecord(tbl) then begin
      Result := TPendingOrphan(Items[i]);
      break;
    end;
end;

procedure  TPendingOrphans.AddRecord(tbl: tTable; aPriIXFields: tStringList);
var i: integer;
    p: TPendingOrphan;
begin
  p := TPendingOrphan.Create(tbl,aPriIXFields);
  for i := 0 to count-1 do
    if TPendingOrphan(Items[i]).CompareInstance(p) then begin
      TPendingOrphan(Items[i]).PendingInPass := CurrentPassNumber; { failed to delete orphan in the current pass }
      p.free;
      exit;
    end;
  Add(p);  
end;

{ TBDEUtil }

constructor TBDEUtil.Create;
begin
  Check(TUInit({$IFDEF WIN32}vhtSes{$ELSE}@vhtSes{$ENDIF}));
end;

destructor TBDEUtil.Destroy;
begin
  Check(TUExit(vhtSes));
end;

procedure TBDEUtil.RegisterCallback(Clback: Pointer; clbID: longint);
begin
 Check(DbiRegisterCallBack(nil, cbGENPROGRESS, clbID,
            sizeof(TUVerifyCallBack), @CbInfo,
{$IFDEF WIN32}
            ClBack));
{$ELSE}
            pfDBICallBack(ClBack)));
{$ENDIF}
end;

procedure TBDEUtil.UnRegisterCallback(clbID: longint);
begin
  Check(DbiRegisterCallBack(nil, cbGENPROGRESS, clbID,
           sizeof(TUVerifyCallBack), @CbInfo, nil));
end;

{ TRbldClass1 }

Constructor TRbldClass1.Create;
begin
  TabVerify := TTable.Create(NIL);
  TabVerify.Exclusive := True;
  TabVerify.databasename := session.privatedir;
  TabVerify.tablename := verifydb;
  BDEUtil := TBDEUtil.Create;
end;

Destructor TRbldClass1.Destroy;
{ delete all auxiliary tables }
begin
  BDEUtil.Free;
  TabVerify.Close;
  Try
    TabVerify.tablename := verifydb;
    TabVerify.Deletetable;
  Except
  End;
  Try
    TabVerify.tablename := backupdb;
    TabVerify.Deletetable;
  Except
  End;
  Try
    TabVerify.tablename := rebuilddb;
    TabVerify.Deletetable;
  Except
  End;
  Try
    TabVerify.tablename := keyvioldb;
    TabVerify.Deletetable;
  Except
  End;
  TabVerify.Free;
end;

Function TRbldClass1.VerifyTable(Const Table: String; Var Msg: Integer; const pswd: string): DBIResult;
var
  psztable,pverifydb: array [0..max_Path] of Char;
  ppswd  : array [0..MaxPasswLen] of Char;
{$IFNDEF WIN32}
  wmsg: word absolute Msg;
{$ENDIF}
begin
  Check(TUExit(BDEUtil.vHtSes));
  Check(TUInit({$IFDEF WIN32}BDEUtil.vHtSes{$ELSE}@BDEUtil.vHtSes{$ENDIF}));
  Msg := 0;
  StrPCopy(ppswd,pswd);
  StrPCopy(psztable,Table);
  StrPCopy(pverifydb,_forcebackslash(session.privateDir)+verifydb);
  if pswd='' then
    Result := TUVerifyTable(BDEUtil.vHtSes, psztable, szPARADOX, pverifydb, nil, 0,
                          {$IFDEF WIN32}
                              Msg)
                          {$ELSE}
                              wMsg)
                          {$ENDIF}
  else
    Result := TUVerifyTable(BDEUtil.vHtSes, psztable, szPARADOX, pverifydb, ppswd, 0,
                          {$IFDEF WIN32}
                              Msg);
                          {$ELSE}
                              wMsg);
                          {$ENDIF}
end;


Function TRbldClass1.Rebuildtable(Const szTable,pswd: String): DBIResult;
var
  iFld, iIdx, iSec, iVal, iRI, iOptP, iOptD: word;
  TblDesc: CRTBlDesc;
  pszTable,
  pBackupdb,
  pRebuilddb,
  pKeyvioldb,
  pProblemdb: array [0..max_Path] of Char;
  ppswd    : array [0..MaxPasswLen] of Char;
begin
  Check(TUExit(BDEUtil.vHtSes));
  Check(TUInit({$IFDEF WIN32}BDEUtil.vHtSes{$ELSE}@BDEUtil.vHtSes{$ENDIF}));
  StrPCopy(ppswd,pswd);
  StrPCopy(pszTable,szTable);
  StrPCopy(pBackupdb, _forcebackslash(session.privateDir)+backupdb);
  StrPCopy(pKeyvioldb,_forcebackslash(session.privateDir)+keyvioldb);
  StrPCopy(pProblemdb,_forcebackslash(session.privateDir)+problemdb);
  StrPCopy(pRebuilddb,_forcebackslash(session.privateDir)+rebuilddb);
  Result := TUGetCRTblDescCount(BDEUtil.vhTSes, psztable, iFld, iIdx, iSec, iVal, iRI,
                                iOptP, iOptD);
  if Result = DBIERR_NONE then begin
    FillChar(TblDesc, SizeOf(CRTBlDesc), 0);
    StrPCopy(TblDesc.szTblType   , szParadox);
    StrCopy (TblDesc.szTblName   , pszTable);
    StrCopy (TblDesc.szErrTblName, pRebuilddb);

    TblDesc.iFldCount := iFld;
    GetMem(TblDesc.pFldDesc, (iFld * SizeOf(FldDesc)));

    TblDesc.iIdxCount := iIdx;
    GetMem(TblDesc.pIdxDesc, (iIdx * SizeOf(IdxDesc)));

    TblDesc.iSecRecCount := iSec;
    GetMem(TblDesc.pSecDesc, (iSec * SizeOf(SecDesc)));

    TblDesc.iValChkCount := iVal;
    GetMem(TblDesc.pvchkDesc, (iVal * SizeOf(VCHKDesc)));

    TblDesc.iRintCount := iRI;
    GetMem(TblDesc.printDesc, (iRI * SizeOf(RINTDesc)));

    TblDesc.iOptParams := iOptP;
    GetMem(TblDesc.pfldOptParams, (iOptP * sizeOf(FLDDesc)));

    GetMem(TblDesc.pOptData, (iOptD * DBIMAXSCFLDLEN));
    try
      if pswd='' then
        Result := TUFillCRTblDesc(BDEUtil.vhTSes, @TblDesc, psztable, nil)
      else
        Result := TUFillCRTblDesc(BDEUtil.vhTSes, @TblDesc, psztable, ppswd);
      if Result = DBIERR_NONE then begin
        Result := TURebuildTable(BDEUtil.vhTSes, pszTable, szPARADOX,
                   pBackupdb, pKeyvioldb,pProblemdb, @TblDesc);
      end;
    finally
      if TblDesc.pFldDesc <> NIL then
      FreeMem(TblDesc.pFldDesc, (iFld * SizeOf(FldDesc)));
      if TblDesc.pIdxDesc <> NIL then
      FreeMem(TblDesc.pIdxDesc, (iIdx * SizeOf(IdxDesc)));
      if TblDesc.pSecDesc <> NIL then
      FreeMem(TblDesc.pSecDesc, (iSec * SizeOf(SecDesc)));
      if TblDesc.pvchkDesc <> NIL then
      FreeMem(TblDesc.pvchkDesc, (iVal * SizeOf(VCHKDesc)));
      if TblDesc.printDesc <> NIL then
      FreeMem(TblDesc.printDesc, (iRI * SizeOf(RINTDesc)));
      if TblDesc.pfldOptParams <> NIL then
      FreeMem(TblDesc.pfldOptParams, (iOptP * sizeOf(FLDDesc)));
      if TblDesc.pOptData <> NIL then
      FreeMem(TblDesc.pOptData, (iOptD * DBIMAXSCFLDLEN));
    end;
  end;
end;

{$IFDEF WIN32}
procedure TRbldClass1.NoDefaultpassword(Sender: TObject; var Continue: Boolean);
{$ELSE}
procedure NoDefaultpassword(Sender: TObject; var Continue: Boolean); far;
{$ENDIF}
{ OnPassword event handler for default session that disables default password dialog.
  Passwords can come through MyGetpassword routine only }
begin
  Continue := false;
end;


{ CheckAndRepairTables procedure }

Procedure CheckAndRepairTables(
               const datBname: string;
               IsAlias: boolean;
               scopeRepair: integer;
               DoPack: Boolean;
               TblCallback,
               DatabaseCallback: pointer;
               const tableSelected:String;
               SelectDBFolderFunc:TSelectDBFolderFunc);

     procedure DoTableCallback(percentdone : smallint; amsg : string);
     var
       ProgressDesc: CBPROGRESSDesc;
     begin { DoTableCallback - invokes TABLE progress callback (if assigned) }
       if assigned(TblCallback) then begin
         ProgressDesc.iPercentdone := percentdone;
         StrPCopy(ProgressDesc.szMsg,aMsg);
         if TGenProgressCallBack(TblCallback)(cbGenProgress,clbdataPDXRBLD,@ProgressDesc)=cbrAbort
           then sysutils.abort;
       end
     end; { DoTableCallback }

     procedure DoDatabaseCallback(percentdone : integer;
                                 aProcess    : TRebuildProcess;
                                 aState      : TCallbackState;
                                 ErrCode     : integer;
                                 ErrMsg      : string;
                             var aRecord: TCallbackRecord);
     begin { DoDatabaseCallback - invokes database progress callback (if assigned) }
       if assigned(DatabaseCallback) then begin
         aRecord.aprocess    := aprocess;
         aRecord.percentdone := percentdone;
         aRecord.aState      := aState;
         aRecord.ErrorCode   := ErrCode;
         aRecord.Errormessage:= ErrMsg;
         if TGenProgressCallBack(DatabaseCallback)(cbGENPROGRESS,clbdataPDXRBLD,@arecord)=cbrAbort
           then sysutils.abort
       end;
     end; { DoDatabaseCallback }

     procedure CheckOrphans(const datafolder,
                            detailtable: string;
                            INIf: tRKSIniFile;
                            VAR OrphansFound,OrphansDeleted: Longint);
     const
         separs = [';'];
     var odeleteconfirm,reccnt,i,j,k: integer;
         mtblstr,
         dtblstr,
         istr,
         mast,
         mfld,
         dfld: string;
         dpriixlist,  { holds detail table primary index field names }
         mfldList,    { holds master table link field names }
         dfldList: tStringList; { holds detail table link field names }
         dtbl,mtbl: ttable;
         RecordDeleted,
         RecordOK,
         IndexSort: boolean;
         values: variant;
         tOrph : tPendingOrphan;

         Procedure RaiseOCheckIniError;
         { Raised if there is an invalid entry in OCheck.INI file - this is always
           programmer's error }
         begin
           raise Exception.CreateFmt(InvalidOCheckINI,[iniOcheck,detailtable,mast,mfld,dfld]);
         end;

     begin { CheckOrphans }
       if INIf = nil then exit;
{$IFNDEF WIN32}
       Values := nil;
{$ENDIF}
       dtblstr := ChangeFileExt(detailtable,'');
       i    := 1;
       dtbl := nil;
       mtbl := nil;
       OrphansFound  := 0;
       OrphansDeleted:= 0;
       dpriixlist := nil;
       mfldList := nil;
       dfldList := nil;
       RecordOK := True;
       odeleteconfirm := mrIgnore;
       Try
         dpriixlist := tStringList.Create;
         mfldList := tStringList.Create;
         dfldList := tStringList.Create;
         repeat
           istr := inttostr(i);
           mast := INIf.ReadString(dtblstr,iniMaster+istr,'');
           mfld := INIf.ReadString(dtblstr,iniMfields+istr,'');
           dfld := INIf.ReadString(dtblstr,iniDfields+istr,'');
           if (mast <> '') and (mfld <> '') and (dfld <> '') then begin
             mtblstr := ChangeFileExt(mast,'');
             mast := mtblstr+'.DB';
             if dtbl = nil then begin
               dtbl := ttable.create(nil);
               dtbl.databasename := datafolder;
               dtbl.tablename := dtblstr+'.DB';
               dtbl.open;
             end;
             reccnt := dtbl.recordcount;
             If reccnt = 0 then break; { avoid unnecessary assignements }
             if dpriixlist.Count=0 then begin { first iteration - only once for current detail table }
               dtbl.IndexDefs.Update;
               For j := 0 to dtbl.IndexDefs.Count-1 do if ixPrimary in dtbl.IndexDefs.Items[j].Options then begin
                 parseWords(dtbl.IndexDefs.Items[j].Fields,separs,dpriixlist);
                 break;
               end;
               if dpriixlist.Count=0 then { Invalid entry in INI file, table has no primary index }
                 Raise Exception.CreateFmt(priIxRequired,[Detailtable]);
             end;
             parseWords(mfld,separs,mfldList);
             parseWords(dfld,separs,dfldList);
             if (mfldList.Count=0) or (mfldList.Count <> dfldList.Count) then { Invalid entry in INI file }
                 RaiseOCheckIniError;
             if (mtbl <> nil) and (mtbl.tablename <> mast) then mtbl.close
             else begin
               mtbl := ttable.create(nil);
               mtbl.databasename := datafolder;
             end;
             mtbl.Indexname := '';
             if not mtbl.active then begin
               mtbl.tablename := mast;
               Try
                 mtbl.open;
               Except  { Invalid entry in INI file }
                 RaiseOCheckIniError;
               End;
             end;
             for j := 0 to mfldList.Count-1 do begin
               if mtbl.FindField(mfldList[j]) = nil then { Invalid entry in INI file }
                 RaiseOCheckIniError;
               if dtbl.FindField(dfldList[j]) = nil then { Invalid entry in INI file }
                 RaiseOCheckIniError;
             end;
             IndexSort := true;
             Try
               Try
                 mtbl.indexfieldnames := mfld;
               Except
                 IndexSort := false;  { Master table index cannot be set, we'll have to use LOCATE }
                 Values := VarArrayCreate([0, mfldList.Count-1]{$IFDEF WIN32},varVariant{$ENDIF});
               End;
               dtbl.first;
               k := 0;
               while not dtbl.eof do begin { Iterate detail table }
                 if reccnt > 0 then DoTableCallback(round(100*k/reccnt),'')
                 else DoTableCallback(100,'');
                 if IndexSort then begin  { Master table is keyed on Link fields }
                   mtbl.setkey;           { .. we can use GotoKey }
                   mtbl.KeyFieldCount := mfldList.Count;
                   Try
                     for j := 0 to mfldList.Count-1 do
                       mtbl[mfldList[j]] := dtbl[dfldList[j]];
                   Except { Invalid entry in INI file }
                     RaiseOCheckIniError;
                   End;
                   RecordOK := mtbl.GoToKey;  { Master exists ? }
                 end else begin { Master table is *not* keyed on Link fields - must use Locate }
                   for j := 0 to mfldList.Count - 1 do Values[j] := dtbl[dfldList[j]];
                   Try
                     RecordOK := mtbl.Locate(mfld,Values,[]); { Master exists ? }
                   Except { Invalid entry in INI file }
                     RaiseOCheckIniError;
                   End;
                 end;
                 RecordDeleted := false;
                 if not RecordOK                                { Detail has no master }
                    or (PendingOrphans.FindRecord(mtbl) <> nil) { or the master is itself an orphan }
                 then begin
                   tOrph := PendingOrphans.FindRecord(dtbl);    { pendingOrphans holds all orphans
                                                                  that we failed to delete in current pass }
                   if (tOrph = nil) or (tOrph.PendingInPass <> CurrentpassNumber) then begin
                    { orphans that we already failed to delete in current pass are skipped }
                     inc(OrphansFound);
                     if scopeRepair <> rsNone then begin
                       if odeleteconfirm = mrIgnore then
                         odeleteconfirm := MessageDlg(Format(LoadStr(msg_Orphanedrecordfoundintable),[detailtable]),
                                   mtConfirmation,[mbYes,mbNo],0);
                       if odeleteconfirm = mrYes Then
                       Try
                         dtbl.delete;
                         dec(reccnt);
                         inc(OrphansDeleted);
                         RecordDeleted := true;
                       Except
                         { Handle all exceptions. Most probable cause is the referential integrity violation }
                       end
                     end;
                     if not RecordDeleted then PendingOrphans.AddRecord(dtbl,dpriixlist);
                   end;
                 end; { if NOT RecordOK or PendingOrphans.. ) }
                 if not RecordDeleted then begin
                   inc(k);
                   dtbl.next;
                 end;
               end; { While not dtbl.EOF }
             Finally
             {$IFNDEF WIN32}
               Values.Free;
               Values := nil;
             {$ENDIF}
             End;
           end else {NOT ((mast <> '') and (mfld <> '') and (dfld <> ''))} break;
           inc(i);
         until false;
       finally
         dpriixlist.free;
         mfldList.Free;
         dfldList.Free;
         if dtbl <> nil then dtbl.close;
         dtbl.free;
         if mtbl <> nil then mtbl.close;
         mtbl.free;
         DoTableCallback(0,'');
       end;
     end; { CheckOrphans }

     Function PerformRestructure(const datbName,Tblname,PswdString: string; TotalTabRecords: Longint;
                const Option,Optdata: String; Processdata: integer): dbiResult;
    { Packs table. TotalTabRecords value is stored into PackProgressDesc record for
      callback purpose }
     Var
       TblDesc:CRTblDesc;
       FDesc: FLDDesc;
       PackhDb: hDBIDb;
       RestrProgressDesc: tCBRESTRPROGRESSDesc;
     begin
       FillChar(TblDesc, SizeOf(CRTblDesc), 0);
       FillChar(RestrProgressDesc,SizeOf(RestrProgressDesc),#0);
       RestrProgressDesc.TotalRecords := TotalTabRecords;
       with session.Finddatabase(datBname) do PackhDb := handle;
       StrPCopy(TblDesc.szTblName, TblName);
       StrPCopy(TblDesc.szTblType,szPARADOX);
       TblDesc.bPack := True;
       TblDesc.bProtected := Length(pswdString) > 0;
       if TblDesc.bProtected then StrPCopy(TblDesc.szPassword,pswdString);
       if Option <> '' then begin
         FillChar(FDesc,SizeOf(FDesc),#0);
         FDesc.iOffset := 0;
         FDesc.iLen := Length(OptData) + 1;
         StrPCopy(FDesc.szName, Option);
         TblDesc.iOptParams := 1;  { Only one optional parameter }
         TblDesc.pFldOptParams := @FDesc;
         TblDesc.pOptData := @OptData[1];
       end;
      { registering callback for restructure process }       
       if TblCallback <> NIL then
         Check(DbiRegisterCallBack(nil, cbGENPROGRESS, Processdata,
           sizeof(RestrProgressDesc), @RestrProgressDesc,
                                          {$IFDEF WIN32}
                                             TblCallback));
                                          {$ELSE}
                                             pfDBICallBack(TblCallback)));
                                          {$ENDIF}
       Try
         Result:= DbiDoRestructure(PackhDB, 1, @TblDesc, nil, nil, nil, False);
       Finally
        { UNregistering callback for restructure process }
         if TblCallback <> NIL then
           Check(DbiRegisterCallBack(nil, cbGENPROGRESS, clbdataPack,
             sizeof(RestrProgressDesc), @RestrProgressDesc, nil));
       end;
     end;

VAR
  XIndexReplaceHeaderList: tStringList; {holds extensions of *.X* indices whose headers should be replaced }
  BorrowFolder: string; { folder containing backup-ed tables from which the header will be borrowed }

     Function BorrowHeader(Const ForDBFile,CorruptionDescr: String; XIndices: boolean): Boolean;
      { Borrowing header from backup table. Functions returns false if either
        SelectDBFolderFunc is unassigned or user cancelled backup folder selection.
        Otherwise returns true, which means that user successfuly selected
        backup folder for header borrow.
        Params:
          ForDBFile - table name for which header(s) should be borrowed
          CorruptionDescr - dbi error message on corrupted file
          XIndices - header should be borrowed for corrupted *.x* files (indices)
                     if false, header should be borrowed for DB file
      }
     Var
        ptr: Pointer;
        SourceFile,
        DestFile,
        CurrExt,
        HeaderBorrowPath: String;
        DoWriteHeader,
        Done: Boolean;
        I: integer;
     begin { BorrowHeader }
       Result := false;
       if (BorrowFolder <> '') or assigned(SelectDBFolderFunc) then begin
         if BorrowFolder <> '' then HeaderBorrowPath := BorrowFolder
         else
           HeaderBorrowPath := SelectDBFolderFunc(ForDBFile,CorruptionDescr);
         if HeaderBorrowPath<>'' then begin
           if not DirectoryExists(HeaderBorrowPath) then begin
             ShowMessage(LoadStr(msg_Youmustenteravaliddatabase));
             exit;
           end;
           HeaderBorrowPath := _ForceBackSlash(HeaderBorrowPath)+ExtractFilename(ForDbFile);
           GetMem(ptr,PxHeaderSize);
           Try
             i := 0;
             if not XIndices then begin
               CurrExt := '.DB';
             end else begin
               CurrExt := XIndexReplaceHeaderList[i];
             end;
             Repeat
             { read file header }
               SourceFile := ChangeFileExt(HeaderBorrowPath,CurrExt);
               DestFile := ChangeFileExt(ForDBFile,CurrExt);
               if (CompareText(SourceFile,DestFile)=0) or not FileExists(SourceFile) then begin
                 ShowMessage(LoadStr(msg_Youmustenteravaliddatabase));
                 exit;
               end;
               with tFileStream.Create(SourceFile,fmOpenRead or fmShareDenyNone) do TRY
                 Read(ptr^,PxHeaderSize);
               Finally
                 Free;
               end;
               DoWriteHeader := false;
               if (TPxHeader3(ptr^).FileVersionId < $05) {< level 4}
            {$IFNDEF WIN32}
               or (TPxHeader3(ptr^).FileVersionId >= $0C) {>= level 7}
            {$ENDIF}
               then
                 ShowMessage(Format(LoadStr(msg_UnSupportedVersion),[SourceFile]))
               else
               if TPxHeader3(ptr^).headerSize <> PxHeaderSize then
                 ShowMessage(Format(LoadStr(msg_InvalidHeaderSize),[SourceFile]))
               else
               if not (TPxHeader3(ptr^).FileType in [0,2,3,5,6,8]) then
                 ShowMessage(Format(LoadStr(msg_NotaDBFile),[SourceFile]))
               else DoWriteHeader := True;
             { write file header }
               if DoWriteHeader then
               with tFileStream.Create(DestFile,fmOpenWrite or fmShareDenyNone) do TRY
                 Write(ptr^,PxHeaderSize);
               Finally
                 Free;
               end else exit;  { result = false }
               if not XIndices then begin
                 Done := true;
               end else begin
                 inc(i);
                 Done := i >= XIndexReplaceHeaderList.Count;
                 if not done then CurrExt := XIndexReplaceHeaderList[i];
               end;
             until done;
           Finally
             FreeMem(ptr,PxHeaderSize);
           End;
           Result := true;
           BorrowFolder := HeaderBorrowPath; { remember backup path path for current database }
         end;
       end;
     end; { BorrowHeader }

     Function GetAutoincFieldName(h: hDBICur): string;
     Var
       Props: CURProps;
       pF,
       pFields: pFLDDesc;
       i: integer;
       MypLdObj: pLDDesc;
       Len: Byte;
       Done: Bool;
     begin
       Result := '';
       Check(DbiGetCursorProps(h, Props));
       pFields := AllocMem(succ(Props.iFields) * sizeof(FLDDesc));
       try
         Check(DbiGetFieldDescs(h, pFields));
         pF := pFields;
         for i := 0 to Props.iFields - 1 do
           if (pF^.iFldType = fldINT32 ) and
           (pF^.iSubType = fldstAUTOINC) then begin
           { Result := Strpas(pF^.szname); }
           { fix for 4.20 - translate names containing foreign characters correctly }
             Check(DbiGetLdObj(H, Pointer(MypLdObj)));
             Len:= StrLen(pF^.szname);
           {$IFDEF WIN32}
             SetLength(Result, Len);
             Check(DbiNativeToAnsi(MypLdObj, PChar(Result), pF^.szname, Len, Done));
           {$ELSE}
             Result[0] := Char(Len);
             Check(DbiNativeToAnsi(MypLdObj, @Result[1], pF^.szname, Len, Done));
           {$ENDIF}
             break;
           end else inc(pF);
       finally
        FreeMem(pFields, succ(Props.iFields) * sizeof(FLDDesc));
       end;
     end;

LABEL
  RetryAfterHeaderBorrow;
VAR
  packResult,                          { dbiResult after pack }
  VerifyResult,                        { dbiResult after Verifytable }
  RebuildResult : DBIResult;           { dbiResult after Rebuildtable }
  VerifyMsg,                           { Error level after Verifytable }
  RecCnt,                              { Table record count }
  MaxAutoinc,                          { Maximum autoincrement value }
  i,j: integer;                        { Counters }
{$IFDEF WIN32}
  MyPx3PProtectCalled: boolean;        { MyPx3PProtect successfuly called }
  Px3PCounter: integer;                {.. another counter}
  Px3PkeyString: String;               { Current Px3P key }
{$ENDIF}
  XIndexHeaderReplaceMsg,              { Caption for borrow header dialog }
  PswdString,                          { Password }
  Msg,                                 { error messages }
  AutoIncName,                         { Fieldname of autoinc field }
  DBpath,                              { Path to .DB file }
  ValPath,                             { Path to .VAL file }
  DatabasePath,                        { Path to database }
  NonFamilyFile,                       { offending file not belonging to the Paradox file family}
  AllPurposeS: string;                 { string for various purposes }
  TabList,                             { List holding table names }
  EmptyIndexList : tStringList;        { List holding paths to index files that should be truncated }
  FormattedMemoPresent,                { Formatted memo present? }
  AutoInc_in_IX,                       { Autoinc field in index? }
  XHeaderAlreadyBorrowed,              { Index header already borrowed? }
  HeaderAlreadyBorrowed,               { DB header already borrowed? }
  VerifyCompleted,                     { VerifyTable passed? }
  tableIsOK,                           { Table healthy? }
  CanRebuild,                          { Can I rebuild the table? }
  ValExists: Boolean;                  { .VAL file exists? }
  DB_Header: T_PxHeader;               { DB file header class }
  TabError,                            { Table corruption error code }
  ValSync : byte;                      { .DB-.VAL sync byte value }
  TabErrors  : set of byte;            { All table corruption error codes }
  DBFileRecord,                        { Record for callback purposes }
  VerifyRecord: TCallbackRecord;       { dtto }
  OrphansFound,                        { Number of orphans detected }
  OrphansDeleted,                      { Number of orphans deleted }
  PasswChkSum: longint;                { Checksum of master password }
  round100mulIdivListCount: word;      { Precalculated percent-done value for callback purposes }
  OrphansINI : tRKSINIFile;               { INI file containing master-detail links descr. }
  Rbld: tRbldClass1;                   { Class performing Verify/Rebuild }
begin  { CheckAndRepairTables }
  SetPrivDir; { session.Private directory becomes windows\temp\Pdxrbld }
  TabList                 := nil;
  EmptyIndexList          := nil;
  OrphansINI              := nil;
  XIndexReplaceHeaderList := nil;
  Rbld                    := nil;
  DB_Header               := nil;
  VerifyResult            := DBIERR_NONE;
  RebuildResult           := DBIERR_NONE;
  PasswChkSum             := 0;
  MaxAutoinc              := 0;
  FillChar(DBFileRecord,SizeOf(DBFileRecord),#0);
  FillChar(VerifyRecord,SizeOf(VerifyRecord),#0);
  TRY
  { routine initialization }
    XIndexReplaceHeaderList := tStringList.Create;
    EmptyIndexList          := tStringList.Create;
    tabList                 := tStringList.Create;
    DatabasePath := datBname;
    if IsAlias then begin     { Find physical path for alias }
      try
        session.GetAliasParams(datBname,tablist);
        Databasepath := tablist.values['PATH'];
      finally
        tablist.clear;
      end;
    end;
   { Check whether database is in use. If so, raise an exception }
    CheckDatabaseInUse(DatabasePath);
   { Check whether this is repeated pass for the same database.
     If not, clear the list of pending orphans }
    if AnsiCompareText(Databasepath,LastDataFolder) <> 0 then begin
      LastDataFolder := Databasepath;
      BorrowFolder := '';
      PendingOrphans.ClearItems;
      CurrentpassNumber := 1;
    end else
      inc(CurrentpassNumber);
    AllPurposeS := _Forcebackslash(DatabasePath)+iniOcheck;
    if FileExists(AllPurposeS) then OrphansINI := tRKSINIFile.Create(AllPurposeS);
    if tableSelected <> '' then
      tablist.Add(tableSelected)
    else
      Session.GetTableNames(datBname, '*.DB', True {Extensions}, False {SystemTables}, TabList);
    if tabList.Count = 0 then raise Exception.CreateRes(msg_TherearenoParadoxtables);
    Rbld := tRbldClass1.Create;
    Session.OnPassword := {$IFDEF WIN32}Rbld.NoDefaultpassword{$ELSE}NoDefaultpassword{$ENDIF};
    For i := 1 to tabList.Count do
    TRY
     { table dependent variables initialization }
      HeaderAlreadyBorrowed    := false;
      XHeaderAlreadyBorrowed   := false;
      round100mulIdivListCount := round(100*i/tabList.Count);
      NonFamilyFile            := '';
RetryAfterHeaderBorrow:  { because after borrow the header might be out of date }
      XIndexReplaceHeaderList.Clear;
      EmptyIndexList.Clear;
      VerifyCompleted := True;
      tableIsOK       := True;
      CanRebuild      := True;
      FormattedMemoPresent := False;
      PswdString      := '';
      VerifyMsg       := 0;
      TabErrors       := [];
{$IFDEF WIN32}
      MyPx3PProtectCalled := False;
      Px3PkeyString   := '';
      Finalize(DBFileRecord);
      Finalize(VerifyRecord);
{$ENDIF}
      FillChar(DBFileRecord,SizeOf(DBFileRecord),#0);
      FillChar(VerifyRecord,SizeOf(VerifyRecord),#0);
     { Get DB file information for callback routine }
      DBPath := _Forcebackslash(DatabasePath)+tabList.Strings[i-1];
      With DBFileRecord do begin
        Drive     := ExtractFileDrive(DBPath);
        Directory := ExtractFileDir(DBPath);
        tablename := ChangeFileExt(ExtractFileName(DBPath),'');
        Extension := ExtractFileExt(DBPath);
        date      := FiledateTodateTime(FileAge(DBPath));
        Time      := Frac(date);
        date      := trunc(date);
      end;
     { read DB file header }
      Try
        DB_Header := T_PxHeader.Create(DBPath);
      Except
        On EInvalidHeader do begin
          {invalid header , DB_Header remains NIL }
        end;
      End;
{ ----------------------- }
{ Starting verify block 1 }
{ ----------------------- }
      DoDatabaseCallback(round100mulIdivListCount,
                         cpVerify,
                         cstStarting,
                         0,'',
                         DBFileRecord);
     { Password and table level checking }
      if DB_Header <> NIL then begin
        if (DB_Header.TableLevel >= 40) {level 4}
       {$IFNDEF WIN32}
        and (DB_Header.TableLevel < 70) {level 7}
       {$ELSE}
        and (DB_Header.TableLevel < unsupportedlevel) {level > 7}
       {$ENDIF}
        then PasswChkSum := DB_Header.MasterPasswordChecksum { Pdox ver 4 to 7(5) }
        else begin { unsupported version }
          DoDatabaseCallback(round100mulIdivListCount,
                             cpVerify,
                             cstReportingResult,
                             ErrCodeUnsupportedVersion,LoadStr(msg_Tablecannotbeverified)+
                                          Format(LoadStr(msg_UnSupportedVersion),[DBPath]),
                             DBFileRecord);
          Continue; { skip current table }
        end;
        if PasswChkSum <> 0 then begin { table is password protected }
{$IFDEF WIN32}
          For Px3PCounter := 0 to Px3PKeysList.Count - 1 do begin { Px3PKeysList has at least 1 item }
            j := tStringList(Px3PKeysList.Objects[Px3PCounter]).IndexOfObject(tObject(PasswChkSum));
            if j >= 0 then begin
              Px3PKeyString := Px3PKeysList[Px3PCounter];
              PswdString := tStringList(Px3PKeysList.Objects[Px3PCounter]).Strings[j];
              break;
            end;
          end;
          if j < 0 then begin { .. did not found password in the list }
           { prompt for password }
            if MyGetpassword(DBFileRecord.tablename,Px3PKeysList,Px3PKeyString,PswdString,PasswChkSum) then
              AddSessionPassword(Px3PKeyString,PswdString)
            else
              continue; { skip current table }
          end;
          if Px3PKeyString <> '' then begin
            MyPx3PProtect(Px3PKeyString);
            MyPx3PProtectCalled := True;
          end;
{$ELSE}
          j := PswdList.IndexOfObject(tObject(PasswChkSum)); { look in the passwords list }
          if j < 0 then begin { .. did not found password in the list }
            if MyGetpassword(DBFileRecord.tablename,PswdString,PasswChkSum) then { prompt for password }
              AddSessionPassword(PswdString)
            else
              continue; { skip current table }
          end else PswdString := PswdList.Strings[j];
{$ENDIF}
        end;
        { Check (and repair) "VAL file out of date" error }
        ValPath := ChangeFileExt(DBPath, '.VAL');
        ValExists := FileExists(ValPath);
        if ValExists then with tFileStream.Create(ValPath,fmOpenRead or fmShareDenyNone) do TRY
          Read(ValSync,1);
        Finally
          Free;
        end;
        { We must check for and repair the "VAL file out of date" BEFORE any other check is performed,
          because otherwise the table would not open at all : }
        if ValExists and (DB_Header.ValSyncByte <> ValSync) then begin
        { VAL file out of date }
          tableIsOK := false;
          DoDatabaseCallback(round100mulIdivListCount,
                           cpVerify,
                           cstReportingErrors,
                           ErrCodeVALOutOfDate,LoadStr(msg_VALfileoutofdate),
                           DBFileRecord);
          if scopeRepair <> rsNone then begin { Write correct Sync value to the VAL file: }
  { ----------------------- }
  { Starting repair block 1 }
  { ----------------------- }
            DoDatabaseCallback(round100mulIdivListCount,
                           cpRepair,
                           cstStarting,
                           0,'',
                           DBFileRecord);
            with tFileStream.Create(ValPath,fmOpenWrite or fmShareDenyNone) do TRY
              VALSync := DB_Header.ValSyncByte;
              Write(VALSync,1);
              tableIsOK := true;
            Finally
              Free
            end;
            DoDatabaseCallback(round100mulIdivListCount,
                           cpRepair,
                           cstReportingResult,
                           0,LoadStr(msg_VALfilefixed),
                           DBFileRecord);
          end;
        end;
      end; { if DB_Header <> nil }
{ ------------------------------------------------- }
{ Let the flow go further  even with nil DB_Header. }
{ Eventually the header error should be repaired    }
{ by RebuildTable                                   }
{ ------------------------------------------------- }
{ Starting verify block 2 }
{ ----------------------- }
      DoDatabaseCallback(round100mulIdivListCount,
                         cpVerify,
                         cstStarting,
                         0,'',
                         DBFileRecord);
{ Check by using tUtility }
      if tableIsOK then begin
        Msg             := '';
        if TblCallback <> NIL then Rbld.BDEUtil.RegisterCallBack(TblCallback,clbdataTUTILITYVerify);
        Try
          Try
            if DB_Header = nil then begin
              VerifyResult := DBIERR_NONE;
              VerifyMsg    := 1;
            end else
            if ScopeRepair = rsAllSkipTUVerify then begin
              VerifyResult := DBIERR_NONE;
              VerifyMsg    := 0;
            end else
              VerifyResult := Rbld.VerifyTable(DBPath,VerifyMsg,pswdString);
          Except
            On E: Exception do begin
              VerifyResult := ERRCODE_OSUNKNOWN;
              msg := E.message;
            end
          end;
        Finally
          if TblCallback <> NIL then Rbld.BDEUtil.UnRegisterCallBack(clbdataTUTILITYVerify);
        End;
        VerifyCompleted := VerifyResult=DBIERR_NONE;
        if VerifyCompleted then
          DoDatabaseCallback(round100mulIdivListCount,
                           cpVerify,
                           cstReportingResult,
                           DBIERR_NONE,'',
                           DBFileRecord)
          else begin
            if Msg='' then Msg := GetBDEErrorMsg(VerifyResult);
            DoDatabaseCallback(round100mulIdivListCount,
                           cpVerify,
                           cstReportingResult,
                           VerifyResult,LoadStr(msg_Tablecannotbeverified)+Msg,
                           DBFileRecord);
        end;
        if VerifyCompleted then begin
          if VerifyMsg <> 0 { there are errors } then begin
            tableIsOK := false;
           { Getting errors codes }
            if DB_Header = nil then begin { Header inconsistent }
              VerifyRecord := DBFileRecord;
              Include(TabErrors,HeaderInconsistent);
              DoDatabaseCallback(round100mulIdivListCount,
                          cpVerify,
                          cstReportingErrors,
                          HeaderInconsistent,
                          GetTUTILITYString(HeaderInconsistent),
                          VerifyRecord);
            end else
            with Rbld.TabVerify do begin
              Open;
              Try
                While not Eof do begin
                  With VerifyRecord do begin
                    Drive := FieldByname(GetTUTILITYString(TabVerifyDrive)).asString;
                    Directory := FieldByname(GetTUTILITYString(TabVerifyDirectory)).asString;
                    Tablename := FieldByname(GetTUTILITYString(TabVerifyTable_name)).asString;
                    Extension := FieldByname(GetTUTILITYString(TabVerifyExtension)).asString;
                    date :=FieldByname(GetTUTILITYString(TabVerifyDate)).asDateTime;
                    Time :=FieldByname(GetTUTILITYString(TabVerifyTime)).asDateTime;
                  end;
                  TabError := FieldByname(GetTUTILITYString(TabVerifyError_code)).asInteger;
                  if TabError = FormattedMemoError then begin
                    FormattedMemoPresent := True;
                    Next;
                    Continue; {continue While not eof.. Do not include FormattedMemoError into TabErrors }
                  end;
                 { Ver 3.30: if there are files in the database folder that do not belong to
                   Paradox file family, yet their name is table_name.X**, TUTILITY incorrectly
                   assumes that such files are corrupted indices (typically .XLS files).
                   We have to perform check for Paradox family membership ourselves: }
                  if (length(VerifyRecord.Extension) >= 2) and
                     (VerifyRecord.Extension[2] in ['x','X']) then begin
                    { copy extension part after 'X' letter }
                    AllPurposeS := Copy(VerifyRecord.Extension,3,length(VerifyRecord.Extension)-2);
                    if (length(AllPurposeS) < 2) {indices always have have 3 letter extension}
                        or
                       not (AllPurposeS[2] in ['0'..'F']) {last letter is always hexadecimal digit}
                    then begin
                      NonFamilyFile := VerifyRecord.Tablename+VerifyRecord.Extension;
                      Next;
                      Continue; {continue While not eof..  Ignore the "error" }
                    end;
                  end;
                  Include(TabErrors,TabError);
               { Rebuildtable will reconstruct indices if we create empty index files.
                 Just remember index paths now, they will be created later }
                  if TabError = MissingPXFile then
                    EmptyIndexList.Add(ChangeFileExt(DBPath,'.PX'));
                  if TabError = MissingYFile then begin
                    { tUtility reports missing .Y* file. The file reported in Verify table
                      is the associated .X* file - we need to change extension }
                    AllPurposeS := VerifyRecord.Extension;
                    AllPurposeS[2] := 'Y';
                    EmptyIndexList.Add(ChangeFileExt(DBPath,AllPurposeS));
                  end;
                 { Check for index header corruption : }
                  if (UpCase(VerifyRecord.Extension[2]) = 'X')
                     and
                     (TabError in XIndexReplaceHeaderErrors) then begin
                    XIndexReplaceHeaderList.Add(VerifyRecord.Extension);
                    XIndexHeaderReplaceMsg :=
                             FieldByname(GetTUTILITYString(TabVerifyError_message)).asString;
                  end;
                  DoDatabaseCallback(round100mulIdivListCount,
                           cpVerify,
                           cstReportingErrors,
                           TabError,
                           FieldByname(GetTUTILITYString(TabVerifyError_message)).asString,
                           VerifyRecord);
                  Next;
                end; { While not EOF }
              Finally
                Close;
              End;  { Try }
            end { With TabVerify };
            if TabErrors = [] then begin
              tableIsOK := True; { This ignores formatted memo error and other "pseudo" errors }
            end;
          end { if verifyMsg <> 0 }
        end {if verifyCompleted} else tableIsOK := false;
        CanRebuild := (TabErrors*UnreparableErrors = [])
                       and
                     { table will be unreparable if it contains errors that
                       should be fixed by restructure only mixed with other
                       types of errors: }
                      (
                       ((TabErrors*FixByRestructureErrors)= [])
                        or
                       ((TabErrors*FixByRestructureErrors)= TabErrors)
                        or
                       (TabErrors=RuleExceptionCombination1)
                      )
      end;
{ Repair by using tUtility / possibly also perform header borrow }
      if VerifyCompleted and (scopeRepair <> rsNone) and ((scopeRepair >= rsAll) or not TableIsOK) then begin
{ ----------------------- }
{ Starting repair block 2 }
{ ----------------------- }
        DoDatabaseCallback(round100mulIdivListCount,
                           cpRepair,
                           cstStarting,
                           0,'',
                           DBFileRecord);
        Msg           := '';
       { Try to replace headers in corrupted maintained indices }
        if XIndexReplaceHeaderList.Count > 0 then begin
          if not XHeaderAlreadyBorrowed { We try only once }
          and
          BorrowHeader(DBPath,XIndexHeaderReplaceMsg,True) then begin
            DoDatabaseCallback(round100mulIdivListCount,
                         cpRepair,
                         cstReportingResult,
                         0,LoadStr(msg_HeaderBorrowed),
                         DBFileRecord);
            XHeaderAlreadyBorrowed := True;
            GOTO RetryAfterHeaderBorrow;
          end else begin
            CanRebuild := false;
          end;
        end;
        if not CanRebuild then begin
          DoDatabaseCallback(round100mulIdivListCount,
                         cpVerify,
                         cstReportingErrors,
                         ErrCodeUnreparableError,
                         LoadStr(msg_Unreparableerrors),
                         DBFileRecord);
          Continue;
        end;
        Try
         { Some corruptions should be fixed by simple restructure instead of TUTILTY rebuild routine: }
          if (TabErrors <> []) and (TabErrors <= FixByRestructureErrors) then begin
         { previous version:
           "if TabErrors <= FixByRestructureErrors then begin" - corrected for Pdxrbld 3.10
           Previous version just packed uncorrupted tables instead of trying to rebuild them
           with Rebuildtable routine
         }
            RebuildResult := PerformRestructure(datbName,
                                         tabList.Strings[i-1],
                                         PswdString,
                                         DB_Header.CountRecords,
                                         '','',clbdataPack);
          end else begin
            {Refuse rebuild if table contains formatted memo: }
            if FormattedMemoPresent then raise Exception.CreateRes(msg_CannotRebuildFormattedmemo);
            {Refuse rebuild if there are files that TUTILITY incorrectly recognizes as corrupted
             indices. They would be erased by Rebuildtable routine, therefore ask the user to
             remove these files before rebuilding the table }
            if NonFamilyFile <> '' then
              raise Exception.CreateResFmt(msg_CannotRebuildNonFamilyFile,[NonFamilyFile]);
            if TblCallback <> NIL then Rbld.BDEUtil.RegisterCallBack(TblCallback,clbdataTUTILITYRebuild);
            Try
             { Create empty indices }
              For j := 0 to EmptyIndexList.Count-1 do begin
                with tFileStream.Create(EmptyIndexList.Strings[j],fmCreate) do free;
              end;   
              RebuildResult := Rbld.Rebuildtable(DBPath,pswdString);
              {RebuildResult=DBIERR_HEADERCORRUPT means DB/Index file header corrupted }
              {!!! fix in ver 3.40 - restore tail with invariant fld numbers & auxiliary passwords information : }
              if DB_Header <> nil then begin
                if DB_Header.HasAuxPasswords then
                  DB_Header.WriteHeaderParts(DBPath,[pxh_cryptInfo,pxh_InvFieldNumbers,pxh_sortorder])
                else
                  DB_Header.WriteHeaderParts(DBPath,[pxh_InvFieldNumbers]);
              end;
            Finally
              if TblCallback <> NIL then Rbld.BDEUtil.UnRegisterCallBack(clbdataTUTILITYRebuild);
            End;
          end;
        Except
          On E: Exception do begin
            RebuildResult := ERRCODE_OSUNKNOWN;
            msg := E.message;
          end
        end;
       { REread DB file header after tampering with the DB file }
       { the header should be OK now ... }
        DB_Header.Free;
        DB_Header := NIL;
        Try
          DB_Header := T_PxHeader.Create(DBPath);
        Except
          On EInvalidHeader do begin
          {... but it is not }
            RebuildResult := DBIERR_HEADERCORRUPT;
          end;
        End;
        tableIsOK := RebuildResult = DBIERR_NONE;
        if tableIsOK then begin
          if TabErrors <> [] then { any errors found in VerifyTable routine }
          DoDatabaseCallback(round100mulIdivListCount,
                         cpRepair,
                         cstReportingResult,
                         0,LoadStr(msg_Errorsfixed),
                         DBFileRecord)
        end else begin
          if Msg='' then Msg := GetBDEErrorMsg({VerifyResult}RebuildResult); {corrected 21.9.1998 }
          DoDatabaseCallback(round100mulIdivListCount,
                         cpRepair,
                         cstReportingResult,
                         RebuildResult,LoadStr(msg_Tablecannotberepaired)+Msg,
                         DBFileRecord);
    { Header borrow }
          If ( { "Header corrupt" type errors: }
              (RebuildResult=DBIERR_HEADERCORRUPT) or
              (RebuildResult=DBIERR_INVALIDFIELDNAME)
             )  and
          not HeaderAlreadyBorrowed { We try only once } and BorrowHeader(DBPath,msg,false) then begin
            DoDatabaseCallback(round100mulIdivListCount,
                         cpRepair,
                         cstReportingResult,
                         0,LoadStr(msg_HeaderBorrowed),
                         DBFileRecord);
            HeaderAlreadyBorrowed := True;
            GOTO RetryAfterHeaderBorrow;
            { Repeat check/repair as borrowed header might be out of date }
          end;
        end;
      end;
{ Check (and repair) corrupted autoinc generator
  tUtility does not recognize corrupted autoinc generator
  .. we have to check ourselves ...                        }
      if VerifyCompleted and tableIsOK then begin
{ ----------------------- }
{ Starting verify block 3 }
{ ----------------------- }
        DoDatabaseCallback(round100mulIdivListCount,
                       cpVerify,
                       cstStarting,
                       0,'',
                       DBFileRecord);
        AutoIncName := '';
        Try
          With ttable.Create(NIL) do TRY
            databasename := datBname;
            tablename := tabList.Strings[i-1];
            Exclusive := True;
            Open; { Table is in default (primary index) order }
            if not EOF {not empty} then begin
              FieldDefs.Update;
              IndexDefs.Update;
              AutoIncName := GetAutoincFieldName(handle);
              if AutoIncName <> '' then begin
               { If some of the table indices contain autoincrement field as the very first field,
                 we can find the maximum value simply by applying this index
                 and going to the last record. Here we check whether this is the case. }
                AutoInc_in_IX := false;
                For j := 0 to IndexDefs.Count-1 do
                if (Pos(AutoIncName,IndexDefs.Items[j].Fields)=1) and
                  (  { still, AutoIncName COULD be substring of some other field name }
                   (length(IndexDefs.Items[j].Fields)=length(AutoIncName)) or
                   (pos(';',IndexDefs.Items[j].Fields)=succ(length(AutoIncName))) {!}
                   ) then begin
                { AutoInc is the first field in the index - most common autoinc usage }
                  AutoInc_in_IX := True;
                  IndexFieldnames := AutoIncName;
                  if ixDescending in IndexDefs.Items[j].Options then
                    First
                  else
                    Last;
                  MaxAutoinc := FieldByname(AutoIncName).AsInteger;
                  break;
                end;
                if not AutoInc_in_IX then begin { Autoinc not used as the first field in any index }
                  First;                        { we have to find max autoinc by iterating the table }
                  MaxAutoinc := FieldByname(AutoIncName).AsInteger;
                  RecCnt := RecordCount;
                  j := 0;
                  While not eof do begin
                    DoTableCallback(round(100*j/reccnt),'');
                    if FieldByname(AutoIncName).AsInteger > MaxAutoinc then
                      MaxAutoinc := FieldByname(AutoIncName).AsInteger;
                    inc(j);
                    next;
                  end;
                  DoTableCallback(0,'');
                end;
              end;
            end;
          Finally
            Close;
            Free;
          end;
          DoDatabaseCallback(round100mulIdivListCount,
                       cpVerify,
                       cstReportingResult,
                       0,'',
                       DBFileRecord);
        Except
          On E: Exception do begin
            tableIsOK := false;
            if E is EDBEngineError then With E as EDBEngineError do begin
              DoDatabaseCallback(round100mulIdivListCount,
                       cpVerify,
                       cstReportingResult,
                       Errors[0].Errorcode,LoadStr(msg_Cannotopentable)+E.message,
                       DBFileRecord);
            end else
              DoDatabaseCallback(round100mulIdivListCount,
                       cpVerify,
                       cstReportingResult,
                       ERRCODE_OSUNKNOWN,LoadStr(msg_Cannotopentable)+E.message,
                       DBFileRecord)
          end
        End;
        if tableIsOK and (AutoIncName <> '') { table does have Autoinc field } then begin
          if maxAutoinc > DB_Header.AutoIncLong then begin
            DoDatabaseCallback(round100mulIdivListCount,
                       cpVerify,
                       cstReportingErrors,
                       ErrCodeCorruptedAutoinc,LoadStr(msg_Corruptedautoincrementgenerator),
                       DBFileRecord);
            tableIsOK := false;
           {corrupted autoinc generator}
            if scopeRepair <> rsNone then begin
{ ----------------------- }
{ Starting repair block 3 }
{ ----------------------- }
              DoDatabaseCallback(round100mulIdivListCount,
                       cpRepair,
                       cstStarting,
                       0,'',
                       DBFileRecord);
              DB_Header.AutoIncLong := maxAutoinc;
              { rewrite common header }
              DB_Header.WriteHeaderParts(DBPath,[pxh_3]);
              tableIsOK := True;
              DoDatabaseCallback(round100mulIdivListCount,
                       cpRepair,
                       cstReportingResult,
                       0,LoadStr(msg_Autoincrementgeneratorfixed),
                       DBFileRecord);
            end
          end { if maxAutoinc > DBFileRec.lastauto }
        end; {if tableIsOK and (AutoIncName <> '') }
      end;
{ Orphans checking - does not modify TableIsOK variable because presence of
  orphaned records does not mean table corruption }
      if VerifyCompleted and tableIsOK then begin
        if OrphansINI <> NIL then begin
          AllPurposeS := ChangeFileExt(tabList.Strings[i-1],''); { get rid of '.DB' }
          if OrphansINI.ReadString(AllPurposeS,iniMaster+'1','') <> '' { There are entries for current table }
          then begin
{ ------------------------------ }
{ Starting orphans check/removal }
{ ------------------------------ }
            DoDatabaseCallback(round100mulIdivListCount,
                       cpOrphansCheck,
                       cstStarting,
                       0,'',
                       DBFileRecord);
            Try
              CheckOrphans(DatabasePath,AllPurposeS,OrphansINI, OrphansFound,OrphansDeleted);
              if OrphansFound > 0 then begin
                DoDatabaseCallback(round100mulIdivListCount,
                      cpOrphansCheck,
                      cstReportingErrors,
                      ErrCodeOrphansFound,Format(LoadStr(msg_ThereAreOrphans),[OrphansFound]),
                      DBFileRecord);
                if OrphansDeleted = 0 then begin
                  if scopeRepair <> rsNone then
                  DoDatabaseCallback(round100mulIdivListCount,
                        cpOrphansCheck,
                        cstReportingErrors,
                        ErrCodeOrphansFound,LoadStr(msg_OrphansNotDeleted),
                        DBFileRecord)
                end else if OrphansDeleted = OrphansFound then
                DoDatabaseCallback(round100mulIdivListCount,
                      cpOrphansCheck,
                      cstReportingErrors,
                      ErrCodeOrphansFound,LoadStr(msg_AllOrphansDeleted),
                      DBFileRecord)
                else
                DoDatabaseCallback(round100mulIdivListCount,
                      cpOrphansCheck,
                      cstReportingErrors,
                      ErrCodeOrphansFound,Format(LoadStr(msg_SomeOrphansDeleted),[OrphansDeleted]),
                      DBFileRecord);
              end;
              DoDatabaseCallback(round100mulIdivListCount,
                      cpOrphansCheck,
                      cstReportingResult,
                      DBIERR_NONE,'',
                      DBFileRecord);
          Except
              { handle only database errors }
              On E: EDBEngineError do
                DoDatabaseCallback(round100mulIdivListCount,
                     cpOrphansCheck,
                     cstReportingResult,
                     E.Errors[0].Errorcode,LoadStr(msg_CannotCheckOrphans)+E.message,
                     DBFileRecord);

            End;
          end;
        end;
      end;
{  Pack the table if the user did choose so and there are either no errors or all erros have been fixed }
      if VerifyCompleted and tableIsOK then begin
        if DoPack then begin
{ ------------- }
{ Starting pack }
{ ------------- }
          DoDatabaseCallback(round100mulIdivListCount,
                       cpPack,
                       cstStarting,
                       0,'',
                       DBFileRecord);
          packResult:= PerformRestructure(datbName,
                                   tabList.Strings[i-1],
                                   PswdString,
                                   DB_Header.CountRecords,
                                   '','',clbdataPack);
          if packResult=DBIERR_NONE then
            DoDatabaseCallback(round100mulIdivListCount,
                       cpPack,
                       cstReportingResult,
                       DBIERR_NONE,'',
                       DBFileRecord)
          else
            DoDatabaseCallback(round100mulIdivListCount,
                       cpPack,
                       cstReportingResult,
                       packResult,
                       LoadStr(msg_Unabletopacktable)+GetBDEErrorMsg(packResult),
                       DBFileRecord);
        end { if DoPack };
      end;
      DB_Header.Free;
      DB_Header := NIL;
    Finally
{$IFDEF WIN32}
      if MyPx3PProtectCalled then MyPx3PProtect('')
{$ENDIF}
    End { for all selected tables }
  Finally
    DB_Header.Free;
    TabList.Free;
    XIndexReplaceHeaderList.Free;
    EmptyIndexList.Free;
    OrphansINI.Free;
    Session.OnPassword := NIL;
    Rbld.Free;
  end;
end; { CheckAndRepairTables }

{$IFDEF  WIN32}
Procedure RemoveAllPx3PKeys;
Var i: Integer;
    PswdList: tStringList;
begin
  For i := Px3PKeysList.Count-1 downto 1 do begin { item 0 = '', never remove }
    PswdList := tStringList(Px3PKeysList.Objects[i]);
    Px3PKeysList.Delete(i);
    PswdList.Free;
  end;
end;

Procedure RemovePx3PKey(const Px3PKey: string);
Var i: Integer;
    PswdList: tStringList;
begin
  if Px3PKey = '' then exit; {never remove}
  For i := 0 to Px3PKeysList.Count-1 do
    if Px3PKeysList[i] = Px3PKey then begin
      PswdList := tStringList(Px3PKeysList.Objects[i]);
      Px3PKeysList.Delete(i);
      PswdList.Free;
      break;
    end;
end;

{ With Px3P protection enabled, passwords are associated with Px3P keys.
  Password for tables protected with default Paradox encryption can
  be added as AddSessionPassword('',Password).
  If the database has been protected by using Px3P key, but individual
  table master passwords are not known because they were set by the user,
  Pdxrbld should be prepared for the Px3P key existence by calling
  AddSessionPassword with empty Pwd argument:
       AddSessionPassword(Px3PKey,'') - this adds key Px3PKey into the list
  of known Px3P keys.
  When Pdxrbld detects password protected table, it prompts the user
  for master password. It then iterates the list of known Px3P keys and
  checks whether the password yields the correct password checksum with any of
  the known Px3P keys (password checksum is known as it was extracted from
  the DB file header). If a match is found, Pdxrbld associates this password
  with found Px3P key and calls Px3PProtect(Px3PKey_Found) before checking/
  rebuilding the table.
}
Procedure AddSessionPassword(const Px3PKey,Pwd: string);
{$ELSE}
Procedure AddSessionPassword(const Pwd: string);
{$ENDIF}
Var PasswChkSum: Longint;
{$IFDEF WIN32}
    i: Integer;
    PswdList: tStringList;
{$ENDIF}
begin
{$IFDEF WIN32}
  PswdList := nil;
  For i := 0 to Px3PKeysList.Count-1 do
    if Px3PKeysList[i] = Px3PKey then begin
      PswdList := tStringList(Px3PKeysList.Objects[i]);
      break;
    end;
  if PswdList = nil then begin
    PswdList := tStringList.Create;
    Px3PKeysList.AddObject(Px3PKey,PswdList);
  end;
  if Pwd <> '' then begin
    if Px3PKey <> '' then MyPx3PProtect(Px3PKey);
    Try
      PasswChkSum := CalcPswCheckSum(Pwd);
      if PswdList.IndexOfObject(tObject(PasswChkSum)) < 0 then begin
        Session.Addpassword(Pwd);
        PswdList.AddObject(Pwd,tObject(PasswChkSum));
      end;
    Finally
      if Px3PKey <> '' then MyPx3PProtect('');
    End;
  End;
{$ELSE}
  if Pwd <> '' then begin
    PasswChkSum := CalcPswCheckSum(Pwd);
    if PswdList.IndexOfObject(tObject(PasswChkSum)) < 0 then begin
      Session.Addpassword(Pwd);
      PswdList.AddObject(Pwd,tObject(PasswChkSum));
    end;
  end;
{$ENDIF}
end;

{$IFNDEF WIN32}
var savedExitproc: pointer;
procedure MyExitProc; far;
begin
  ExitProc := savedExitproc;
  PendingOrphans.Free;
  PswdList.Free;
end;
{$ENDIF}

{$IFDEF WIN32}
  procedure DoFinalize;
  var i: integer;
  begin
    PendingOrphans.Free;
    For i := 0 to Px3PKeysList.Count - 1 do
      Px3PKeysList.Objects[i].Free;
    Px3PKeysList.Free;
  end;
{$ENDIF}

initialization
{$IFNDEF WIN32}
  SavedExitproc  := ExitProc;
  ExitProc       := @MyExitProc;
  PswdList       := tStringList.Create;
{$ELSE}
  Px3PKeysList   := tStringList.Create;
  Px3PKeysList.AddObject('',tStringList.Create); { Add password list for default encryption } 
{$ENDIF}
  PendingOrphans := TPendingOrphans.Create;
  LastDataFolder := '';

{$IFDEF WIN32}
finalization
  DoFinalize;
{$ENDIF}

end.

